"""
Keyterm Extraction Utils
------------------------
"""
import itertools
import math
import operator
from decimal import Decimal
from typing import cast, Callable, Collection, Dict, Iterable, List, Optional, Set, Tuple, Union

import numpy as np
from cytoolz import itertoolz
from spacy.tokens import Doc, Span, Token

from .. import extract
from .. import similarity
from .. import utils
from .. import vsm


def normalize_terms(
    terms: Union[Iterable[Span], Iterable[Token]],
    normalize: Optional[Union[str, Callable[[Union[Span, Token]], str]]],
) -> Iterable[str]:
    """
    Transform a sequence of terms from spaCy ``Token`` or ``Span`` s into
    strings, normalized by ``normalize``.

    Args:
        terms
        normalize: If "lemma", lemmatize terms; if "lower", lowercase terms;
            if None, use the form of terms as they appear in ``terms``;
            if a callable, must accept a ``Token`` or ``Span`` and return a str,
            e.g. :func:`textacy.spacier.utils.get_normalized_text()`.

    Yields:
        str
    """
    if normalize == "lemma":
        terms = (term.lemma_ for term in terms)
    elif normalize == "lower":
        terms = (term.lower_ for term in terms)
    elif normalize is None:
        terms = (term.text for term in terms)
    elif callable(normalize):
        terms = (normalize(term) for term in terms)
    else:
        raise ValueError(
            "normalize = {} is invalid; value must be a function or one of {}".format(
                normalize, {"lemma", "lower", None})
        )
    for term in terms:
        yield term


def aggregate_term_variants(
    terms: Set[str],
    *,
    acro_defs: Optional[Dict[str, str]] = None,
    fuzzy_dedupe: bool = True,
) -> List[Set[str]]:
    """
    Take a set of unique terms and aggregate terms that are symbolic, lexical,
    and ordering variants of each other, as well as acronyms and fuzzy string matches.

    Args:
        terms: Set of unique terms with potential duplicates
        acro_defs: If not None, terms that are acronyms will be aggregated
            with their definitions and terms that are definitions will be aggregated
            with their acronyms
        fuzzy_dedupe: If True, fuzzy string matching will be used
            to aggregate similar terms of a sufficient length

    Returns:
        Each item is a set of aggregated terms.

    Notes:
        Partly inspired by aggregation of variants discussed in
        Park, Youngja, Roy J. Byrd, and Branimir K. Boguraev.
        "Automatic glossary extraction: beyond terminology identification."
        Proceedings of the 19th international conference on Computational linguistics-Volume 1.
        Association for Computational Linguistics, 2002.
    """
    agg_terms = []
    seen_terms: Set[str] = set()
    for term in sorted(terms, key=len, reverse=True):
        if term in seen_terms:
            continue
        variants = set([term])
        seen_terms.add(term)

        # symbolic variations
        if "-" in term:
            variant = term.replace("-", " ").strip()
            if variant in terms.difference(seen_terms):
                variants.add(variant)
                seen_terms.add(variant)
        if "/" in term:
            variant = term.replace("/", " ").strip()
            if variant in terms.difference(seen_terms):
                variants.add(variant)
                seen_terms.add(variant)

        # lexical variations
        term_words = term.split()
        # last_word = term_words[-1]
        # # assume last word is a noun
        # last_word_lemmatized = lemmatizer.lemmatize(last_word, 'n')
        # # if the same, either already a lemmatized noun OR a verb; try verb
        # if last_word_lemmatized == last_word:
        #     last_word_lemmatized = lemmatizer.lemmatize(last_word, 'v')
        # # if at least we have a new term... add it
        # if last_word_lemmatized != last_word:
        #     term_lemmatized = ' '.join(term_words[:-1] + [last_word_lemmatized])
        #     if term_lemmatized in terms.difference(seen_terms):
        #         variants.add(term_lemmatized)
        #         seen_terms.add(term_lemmatized)

        # if term is an acronym, add its definition
        # if term is a definition, add its acronym
        if acro_defs:
            for acro, def_ in acro_defs.items():
                if acro.lower() == term.lower():
                    variants.add(def_.lower())
                    seen_terms.add(def_.lower())
                    break
                elif def_.lower() == term.lower():
                    variants.add(acro.lower())
                    seen_terms.add(acro.lower())
                    break

        # if 3+ -word term differs by one word at the start or the end
        # of a longer phrase, aggregate
        if len(term_words) > 2:
            term_minus_first_word = " ".join(term_words[1:])
            term_minus_last_word = " ".join(term_words[:-1])
            if term_minus_first_word in terms.difference(seen_terms):
                variants.add(term_minus_first_word)
                seen_terms.add(term_minus_first_word)
            if term_minus_last_word in terms.difference(seen_terms):
                variants.add(term_minus_last_word)
                seen_terms.add(term_minus_last_word)
            # check for "X of Y" <=> "Y X" term variants
            if " of " in term:
                split_term = term.split(" of ")
                variant = split_term[1] + " " + split_term[0]
                if variant in terms.difference(seen_terms):
                    variants.add(variant)
                    seen_terms.add(variant)

        # intense de-duping for sufficiently long terms
        if fuzzy_dedupe is True and len(term) >= 13:
            for other_term in sorted(
                terms.difference(seen_terms), key=len, reverse=True
            ):
                if len(other_term) < 13:
                    break
                tsr = similarity.token_sort_ratio(term, other_term)
                if tsr > 0.93:
                    variants.add(other_term)
                    seen_terms.add(other_term)
                    break

        agg_terms.append(variants)

    return agg_terms


def get_longest_subsequence_candidates(
    doc: Doc,
    match_func: Callable[[Token], bool],
) -> Iterable[Tuple[Token, ...]]:
    """
    Get candidate keyterms from ``doc``, where candidates are longest consecutive
    subsequences of tokens for which all ``match_func(token)`` is True.

    Args:
        doc
        match_func: Function applied sequentially to each ``Token`` in ``doc``
            that returns True for matching ("good") tokens, False otherwise.

    Yields:
        Next longest consecutive subsequence candidate, as a tuple of constituent tokens.
    """
    for key, words_grp in itertools.groupby(doc, key=match_func):
        if key is True:
            yield tuple(words_grp)


def get_ngram_candidates(
    doc: Doc,
    ns: Union[int, Collection[int]],
    *,
    include_pos: Optional[Union[str, Collection[str]]] = ("NOUN", "PROPN", "ADJ"),
) -> Iterable[Tuple[Token, ...]]:
    """
    Get candidate keyterms from ``doc``, where candidates are n-length sequences
    of tokens (for all n in ``ns``) that don't start/end with a stop word or
    contain punctuation tokens, and whose constituent tokens are filtered by POS tag.

    Args:
        doc
        ns: One or more n values for which to generate n-grams. For example,
            ``2`` gets bigrams; ``(2, 3)`` gets bigrams and trigrams.
        include_pos: One or more POS tags with which to filter ngrams.
            If None, include tokens of all POS tags.

    Yields:
        Next ngram candidate, as a tuple of constituent Tokens.

    See Also:
        :func:`textacy.extract.ngrams()`
    """
    ns = cast(Tuple[int, ...], utils.to_collection(ns, int, tuple))
    include_pos = utils.to_collection(include_pos, str, set)
    ngrams = itertoolz.concat(itertoolz.sliding_window(n, doc) for n in ns)
    ngrams = (
        ngram
        for ngram in ngrams
        if not (ngram[0].is_stop or ngram[-1].is_stop)
        and not any(word.is_punct or word.is_space for word in ngram)
    )
    if include_pos:
        ngrams = (
            ngram
            for ngram in ngrams
            if all(word.pos_ in include_pos for word in ngram)
        )
    for ngram in ngrams:
        yield ngram


def get_pattern_matching_candidates(
    doc: Doc,
    patterns: Union[str, List[str], List[dict], List[List[dict]]],
) -> Iterable[Tuple[Token, ...]]:
    """
    Get candidate keyterms from ``doc``, where candidates are sequences of tokens
    that match any pattern in ``patterns``

    Args:
        doc
        patterns: One or multiple patterns to match against ``doc`` using
            a :class:`spacy.matcher.Matcher`.

    Yields:
        Tuple[:class:`spacy.tokens.Token`]: Next pattern-matching candidate,
        as a tuple of constituent Tokens.

    See Also:
        :func:`textacy.extract.matches()`
    """
    for match in extract.matches(doc, patterns, on_match=None):
        yield tuple(match)


def get_filtered_topn_terms(
    term_scores: Iterable[Tuple[str, float]],
    topn: int,
    *,
    match_threshold: Optional[float] = None,
) -> List[Tuple[str, float]]:
    """
    Build up a list of the ``topn`` terms, filtering out any that are substrings
    of better-scoring terms and optionally filtering out any that are sufficiently
    similar to better-scoring terms.

    Args:
        term_scores: Iterable of (term, score) pairs, sorted in order of score
            from best to worst. Note that this may be from high to low value or low to high,
            depending on the scoring algorithm.
        topn: Maximum number of top-scoring terms to get.
        match_threshold: Minimal edit distance between a term and previously seen terms,
            used to filter out terms that are sufficiently similar
            to higher-scoring terms. Uses :func:`textacy.similarity.token_sort_ratio()`.
    """
    topn_terms = []
    seen_terms: Set[str] = set()
    sim_func = similarity.token_sort_ratio
    for term, score in term_scores:
        # skip terms that are substrings of any higher-scoring term
        if any(term in st for st in seen_terms):
            continue
        # skip terms that are sufficiently similar to any higher-scoring term
        if (
            match_threshold
            and any(sim_func(term, st) >= match_threshold for st in seen_terms)
        ):
            continue
        seen_terms.add(term)
        topn_terms.append((term, score))
        if len(topn_terms) >= topn:
            break
    return topn_terms


def most_discriminating_terms(
    terms_lists: Iterable[Iterable[str]],
    bool_array_grp1: Iterable[bool],
    *,
    max_n_terms: int = 1000,
    top_n_terms: Union[int, float] = 25,
) -> Tuple[List[str], List[str]]:
    """
    Given a collection of documents assigned to 1 of 2 exclusive groups, get the
    ``top_n_terms`` most discriminating terms for group1-and-not-group2 and
    group2-and-not-group1.

    Args:
        terms_lists: Sequence of documents, each as a sequence of (str) terms;
            used as input to :func:`doc_term_matrix()`
        bool_array_grp1: Ordered sequence of True/False values,
            where True corresponds to documents falling into "group 1" and False
            corresponds to those in "group 2".
        max_n_terms: Only consider terms whose document frequency is within
            the top ``max_n_terms`` out of all distinct terms; must be > 0.
        top_n_terms: If int (must be > 0), the total number of most discriminating terms
            to return for each group; if float (must be in the interval (0, 1)),
            the fraction of ``max_n_terms`` to return for each group.

    Returns:
        List of the top ``top_n_terms`` most discriminating terms for grp1-not-grp2, and
        list of the top ``top_n_terms`` most discriminating terms for grp2-not-grp1.

    References:
        King, Gary, Patrick Lam, and Margaret Roberts. "Computer-Assisted Keyword
        and Document Set Discovery from Unstructured Text." (2014).
        http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.458.1445&rep=rep1&type=pdf
    """
    alpha_grp1 = 1
    alpha_grp2 = 1
    if isinstance(top_n_terms, float):
        top_n_terms = int(round(top_n_terms * max_n_terms))
    bool_array_grp1 = np.array(bool_array_grp1)
    bool_array_grp2 = np.invert(bool_array_grp1)

    vectorizer = vsm.Vectorizer(
        tf_type="linear",
        norm=None,
        idf_type="smooth",
        min_df=3,
        max_df=0.95,
        max_n_terms=max_n_terms,
    )
    dtm = vectorizer.fit_transform(terms_lists)
    id2term = vectorizer.id_to_term

    # get doc freqs for all terms in grp1 documents
    dtm_grp1 = dtm[bool_array_grp1, :]
    n_docs_grp1 = dtm_grp1.shape[0]
    doc_freqs_grp1 = vsm.get_doc_freqs(dtm_grp1)

    # get doc freqs for all terms in grp2 documents
    dtm_grp2 = dtm[bool_array_grp2, :]
    n_docs_grp2 = dtm_grp2.shape[0]
    doc_freqs_grp2 = vsm.get_doc_freqs(dtm_grp2)

    # get terms that occur in a larger fraction of grp1 docs than grp2 docs
    term_ids_grp1 = np.where(
        doc_freqs_grp1 / n_docs_grp1 > doc_freqs_grp2 / n_docs_grp2
    )[0]

    # get terms that occur in a larger fraction of grp2 docs than grp1 docs
    term_ids_grp2 = np.where(
        doc_freqs_grp1 / n_docs_grp1 < doc_freqs_grp2 / n_docs_grp2
    )[0]

    # get grp1 terms doc freqs in and not-in grp1 and grp2 docs, plus marginal totals
    grp1_terms_grp1_df = doc_freqs_grp1[term_ids_grp1]
    grp1_terms_grp2_df = doc_freqs_grp2[term_ids_grp1]
    # grp1_terms_grp1_not_df = n_docs_grp1 - grp1_terms_grp1_df
    # grp1_terms_grp2_not_df = n_docs_grp2 - grp1_terms_grp2_df
    # grp1_terms_total_df = grp1_terms_grp1_df + grp1_terms_grp2_df
    # grp1_terms_total_not_df = grp1_terms_grp1_not_df + grp1_terms_grp2_not_df

    # get grp2 terms doc freqs in and not-in grp2 and grp1 docs, plus marginal totals
    grp2_terms_grp2_df = doc_freqs_grp2[term_ids_grp2]
    grp2_terms_grp1_df = doc_freqs_grp1[term_ids_grp2]
    # grp2_terms_grp2_not_df = n_docs_grp2 - grp2_terms_grp2_df
    # grp2_terms_grp1_not_df = n_docs_grp1 - grp2_terms_grp1_df
    # grp2_terms_total_df = grp2_terms_grp2_df + grp2_terms_grp1_df
    # grp2_terms_total_not_df = grp2_terms_grp2_not_df + grp2_terms_grp1_not_df

    # get grp1 terms likelihoods, then sort for most discriminating grp1-not-grp2 terms
    grp1_terms_likelihoods = {}
    for idx, term_id in enumerate(term_ids_grp1):
        term1 = (
            Decimal(math.factorial(grp1_terms_grp1_df[idx] + alpha_grp1 - 1))
            * Decimal(math.factorial(grp1_terms_grp2_df[idx] + alpha_grp2 - 1))
            / Decimal(
                math.factorial(
                    grp1_terms_grp1_df[idx]
                    + grp1_terms_grp2_df[idx]
                    + alpha_grp1
                    + alpha_grp2
                    - 1
                )
            )
        )
        term2 = (
            Decimal(
                math.factorial(n_docs_grp1 - grp1_terms_grp1_df[idx] + alpha_grp1 - 1)
            )
            * Decimal(
                math.factorial(n_docs_grp2 - grp1_terms_grp2_df[idx] + alpha_grp2 - 1)
            )
            / Decimal(
                (
                    math.factorial(
                        n_docs_grp1
                        + n_docs_grp2
                        - grp1_terms_grp1_df[idx]
                        - grp1_terms_grp2_df[idx]
                        + alpha_grp1
                        + alpha_grp2
                        - 1
                    )
                )
            )
        )
        grp1_terms_likelihoods[id2term[term_id]] = term1 * term2
    top_grp1_terms = [
        term
        for term, likelihood in sorted(
            grp1_terms_likelihoods.items(), key=operator.itemgetter(1), reverse=True
        )[:top_n_terms]
    ]

    # get grp2 terms likelihoods, then sort for most discriminating grp2-not-grp1 terms
    grp2_terms_likelihoods = {}
    for idx, term_id in enumerate(term_ids_grp2):
        term1 = (
            Decimal(math.factorial(grp2_terms_grp2_df[idx] + alpha_grp2 - 1))
            * Decimal(math.factorial(grp2_terms_grp1_df[idx] + alpha_grp1 - 1))
            / Decimal(
                math.factorial(
                    grp2_terms_grp2_df[idx]
                    + grp2_terms_grp1_df[idx]
                    + alpha_grp2
                    + alpha_grp1
                    - 1
                )
            )
        )
        term2 = (
            Decimal(
                math.factorial(n_docs_grp2 - grp2_terms_grp2_df[idx] + alpha_grp2 - 1)
            )
            * Decimal(
                math.factorial(n_docs_grp1 - grp2_terms_grp1_df[idx] + alpha_grp1 - 1)
            )
            / Decimal(
                (
                    math.factorial(
                        n_docs_grp2
                        + n_docs_grp1
                        - grp2_terms_grp2_df[idx]
                        - grp2_terms_grp1_df[idx]
                        + alpha_grp2
                        + alpha_grp1
                        - 1
                    )
                )
            )
        )
        grp2_terms_likelihoods[id2term[term_id]] = term1 * term2
    top_grp2_terms = [
        term
        for term, likelihood in sorted(
            grp2_terms_likelihoods.items(), key=operator.itemgetter(1), reverse=True
        )[:top_n_terms]
    ]

    return (top_grp1_terms, top_grp2_terms)



"""
Module for exporting spaCy objects into "third-party" formats.
"""
import collections
import operator

import spacy


def docs_to_gensim(
    spacy_docs,
    spacy_vocab,
    lemmatize=True,
    lowercase=False,
    filter_stops=True,
    filter_punct=True,
    filter_nums=False,
):
    """
    Convert a sequence of ``Doc`` s into a gensim-friendly corpus and a
    string that can be loaded into a :class:`gensim.corpora.Dictionary`.

    Args:
        spacy_docs (Iterable[:class:`spacy.tokens.Doc`])
        spacy_vocab (``spacy.Vocab``)
        lemmatize (bool): if True, use lemmatized strings for words
        lowercase (bool): if True (and ``lemmatize`` is False), use lowercased
            strings for words
        filter_stops (bool): if True, remove stop words from word list
        filter_punct (bool): if True, remove punctuation from word list
        filter_nums (bool): if True, remove numbers from word list

    Returns:
        str: words, their integer ids, and their document frequencies in
        ``spacy_docs``, as a string formatted like `id[TAB]word[TAB]df[NEWLINE]`;
        when written to file, can be converted into a gensim ``Dictionary``
        via :meth:`gensim.corpora.Dictionary.load_from_text()`

        List[List[Tuple[int, int]]]: list of documents as bags-of-words, where
        each doc is a list of (integer word ID, word count) 2-tuples
    """
    count_by = (
        spacy.attrs.LEMMA
        if lemmatize is True
        else spacy.attrs.LOWER
        if lowercase is True
        else spacy.attrs.ORTH
    )
    gcorpus = []
    stringstore = spacy.strings.StringStore()
    doc_freqs = collections.Counter()

    for spacy_doc in spacy_docs:
        bow = (
            (spacy_vocab[tok_id], count)
            for tok_id, count in spacy_doc.count_by(count_by).items()
        )
        bow = ((lex, count) for lex, count in bow if not lex.is_space)
        if filter_stops is True:
            bow = ((lex, count) for lex, count in bow if not lex.is_stop)
        if filter_punct is True:
            bow = ((lex, count) for lex, count in bow if not lex.is_punct)
        if filter_nums is True:
            bow = ((lex, count) for lex, count in bow if not lex.like_num)
        bow = sorted(
            ((stringstore[lex.orth_], count) for lex, count in bow),
            key=operator.itemgetter(0),
        )

        doc_freqs.update(tok_id for tok_id, _ in bow)
        gcorpus.append(bow)

    gdict_str = "\n".join(
        "{}\t{}\t{}".format(i, s, doc_freqs[i])
        for i, s in sorted(enumerate(stringstore), key=operator.itemgetter(1))
    )

    return (gdict_str, gcorpus)


def doc_to_conll(doc):
    """
    Convert a single ``Doc`` into a CoNLL-U formatted str, as described at
    http://universaldependencies.org/docs/format.html.

    Args:
        doc (:class:`spacy.tokens.Doc`)

    Returns:
        str

    Raises:
        ValueError: if ``doc`` is not parsed
    """
    if doc.is_parsed is False:
        raise ValueError("spaCy doc must be parsed")
    rows = []
    for j, sent in enumerate(doc.sents):
        sent_i = sent.start
        sent_id = j + 1
        rows.append("# sent_id {}".format(sent_id))
        for i, tok in enumerate(sent):
            # HACK...
            if tok.is_space:
                form = " "
                lemma = " "
            else:
                form = tok.orth_
                lemma = tok.lemma_
            tok_id = i + 1
            head = tok.head.i - sent_i + 1
            if head == tok_id:
                head = 0
            misc = "SpaceAfter=No" if not tok.whitespace_ else "_"
            rows.append(
                "\t".join(
                    [
                        str(tok_id),
                        form,
                        lemma,
                        tok.pos_,
                        tok.tag_,
                        "_",
                        str(head),
                        tok.dep_.lower(),
                        "_",
                        misc,
                    ]
                )
            )
        rows.append("")  # sentences must be separated by a single newline
    return "\n".join(rows)



"""
Reddit comments
---------------

A collection of up to ~1.5 billion Reddit comments posted from
October 2007 through May 2015.

Records include the following key fields (plus a few others):

    - ``body``: Full text of the comment.
    - ``created_utc``: Date on which the comment was posted.
    - ``subreddit``: Sub-reddit in which the comment was posted, excluding the
      familiar "/r/" prefix.
    - ``score``: Net score (upvotes - downvotes) on the comment.
    - ``gilded``: Number of times this comment received reddit gold.

The raw data was originally collected by /u/Stuck_In_the_Matrix via Reddit's
APIS, and stored for posterity by the `Internet Archive <https://archive.org>`_.
For more details, refer to https://archive.org/details/2015_reddit_comments_corpus.
"""
import itertools
import logging
import pathlib
import re
import urllib.parse
from datetime import datetime
from typing import Iterable, Optional, Set, Tuple, Union

from .. import constants, utils
from .. import io as tio
from .base import Dataset

LOGGER = logging.getLogger(__name__)

NAME = "reddit_comments"
META = {
    "site_url": "https://archive.org/details/2015_reddit_comments_corpus",
    "description": (
        "Collection of ~1.5 billion publicly available Reddit comments "
        "from October 2007 through May 2015."
    ),
}
DOWNLOAD_ROOT = "https://archive.org/download/2015_reddit_comments_corpus/reddit_data/"

RE_REDDIT_LINK = re.compile(r"\[([^]]+)\]\(https?://[^\)]+\)")


class RedditComments(Dataset):
    """
    Stream a collection of Reddit comments from 1 or more compressed files on disk,
    either as texts or text + metadata pairs.

    Download the data (one time only!) or subsets thereof by specifying a date range::

        >>> import textacy.datasets
        >>> ds = textacy.datasets.RedditComments()
        >>> ds.download(date_range=("2007-10", "2008-03"))
        >>> ds.info
        {'name': 'reddit_comments',
         'site_url': 'https://archive.org/details/2015_reddit_comments_corpus',
         'description': 'Collection of ~1.5 billion publicly available Reddit comments from October 2007 through May 2015.'}

    Iterate over comments as texts or records with both text and metadata::

        >>> for text in ds.texts(limit=5):
        ...     print(text)
        >>> for text, meta in ds.records(limit=5):
        ...     print("\\n{} {}\\n{}".format(meta["author"], meta["created_utc"], text))

    Filter comments by a variety of metadata fields and text length::

        >>> for text, meta in ds.records(subreddit="politics", limit=5):
        ...     print(meta["score"], ":", text)
        >>> for text, meta in ds.records(date_range=("2008-01", "2008-03"), limit=5):
        ...     print(meta["created_utc"])
        >>> for text, meta in ds.records(score_range=(10, None), limit=5):
        ...     print(meta["score"], ":", text)
        >>> for text in ds.texts(min_len=2000, limit=5):
        ...     print(len(text))

    Stream comments into a :class:`textacy.Corpus <textacy.corpus.Corpus>`::

        >>> textacy.Corpus("en", data=ds.records(limit=1000))
        Corpus(1000 docs; 27582 tokens)

    Args:
        data_dir: Path to directory on disk under which the data is stored,
            i.e. ``/path/to/data_dir/reddit_comments``. Each file covers a given month,
            as indicated in the filename like "YYYY/RC_YYYY-MM.bz2".

    Attributes:
        full_date_range: First and last dates for which comments
            are available, each as an ISO-formatted string (YYYY-MM-DD).
        filepaths: Full paths on disk for all Reddit comments files
            found under :attr:`ReddictComments.data_dir` directory, sorted
            in chronological order.
    """

    full_date_range: Tuple[str, str] = ("2007-10-01", "2015-06-01")
    _full_score_range: Tuple[int, int] = (-2147483647, 2147483647)

    def __init__(
        self,
        data_dir: Union[str, pathlib.Path] = constants.DEFAULT_DATA_DIR.joinpath(NAME),
    ):
        super().__init__(NAME, meta=META)
        self.data_dir = utils.to_path(data_dir).resolve()
        self._date_range: Optional[Tuple[Optional[str], Optional[str]]] = None

    @property
    def filepaths(self) -> Tuple[str, ...]:
        """
        Full paths on disk for all Reddit comments files found under
        the ``data_dir`` directory, sorted chronologically.
        """
        if self.data_dir.is_dir():
            return tuple(
                sorted(
                    tio.get_filepaths(
                        self.data_dir,
                        match_regex=r"RC_\d{4}",
                        extension=".bz2",
                        recursive=True,
                    )
                )
            )
        else:
            return tuple()

    def download(
        self,
        *,
        date_range: Tuple[Optional[str], Optional[str]] = (None, None),
        force: bool = False,
    ) -> None:
        """
        Download 1 or more monthly Reddit comments files from archive.org
        and save them to disk under the ``data_dir`` directory.

        Args:
            date_range: Interval specifying the [start, end) dates for which
                comments files will be downloaded. Each item must be a str
                formatted as YYYY-MM or YYYY-MM-DD (the latter is converted
                to the corresponding YYYY-MM value). Both start and end values
                must be specified, but a null value for either is automatically
                replaced by the minimum or maximum valid values, respectively.
            force: If True, download the dataset, even if it already
                exists on disk under ``data_dir``.
        """
        date_range = utils.validate_and_clip_range(
            date_range, self.full_date_range, val_type=(str, bytes))
        filestubs = self._generate_filestubs(date_range)
        for filestub in filestubs:
            tio.download_file(
                urllib.parse.urljoin(DOWNLOAD_ROOT, filestub),
                filename=filestub,
                dirpath=self.data_dir,
                force=force,
            )

    def _generate_filestubs(self, date_range):
        """
        Generate a list of monthly filepath stubs in the interval [start, end),
        each with format "YYYY/RC_YYYY-MM.bz2".
        """
        fstubs = []
        start = self._parse_date(date_range[0])
        end = self._parse_date(date_range[1])
        for tot_mo in range(self._total_mos(start) - 1, self._total_mos(end) - 1):
            yr, mo = divmod(tot_mo, 12)
            fstubs.append(datetime(yr, mo + 1, 1).strftime("%Y/RC_%Y-%m.bz2"))
        return tuple(fstubs)

    def _parse_date(self, dt):
        """dt (str) => datetime"""
        try:
            return datetime.strptime(dt, "%Y-%m")
        except ValueError:
            return datetime.strptime(dt, "%Y-%m-%d")

    def _total_mos(self, dt):
        """dt (datetime) => int"""
        return dt.month + 12 * dt.year

    def __iter__(self):
        # for performance reasons, only iterate over files that are requested
        if self._date_range is not None:
            filepaths = [
                self.data_dir.joinpath(filestub)
                for filestub in self._generate_filestubs(self._date_range)
            ]
            for filepath in filepaths:
                if not filepath.is_file():
                    raise OSError(
                        "requested comments file {} not found;\n"
                        "has the dataset been downloaded yet?".format(filepath)
                    )
        else:
            filepaths = self.filepaths
            if not filepaths:
                raise OSError(
                    "no comments files found in {} directory;\n"
                    "has the dataset been downloaded yet?".format(self.data_dir)
                )

        for filepath in filepaths:
            for line in tio.read_json(filepath, mode="rb", lines=True):
                line["created_utc"] = self._convert_timestamp(
                    line.get("created_utc", ""))
                line["retrieved_on"] = self._convert_timestamp(
                    line.get("retrieved_on", ""))
                line["body"] = self._clean_content(line["body"])
                yield line

    def _get_filters(self, subreddit, date_range, score_range, min_len):
        filters = []
        if min_len is not None:
            if min_len < 1:
                raise ValueError("`min_len` must be at least 1")
            filters.append(
                lambda record: len(record.get("body", "")) >= min_len
            )
        if subreddit is not None:
            subreddit = utils.validate_set_members(subreddit, (str, bytes))
            filters.append(
                lambda record: record.get("subreddit") in subreddit
            )
        if date_range is not None:
            date_range = utils.validate_and_clip_range(
                date_range, self.full_date_range, val_type=(str, bytes))
            filters.append(
                lambda record: (
                    record.get("created_utc")
                    and date_range[0] <= record["created_utc"] < date_range[1]
                )
            )
        if score_range is not None:
            score_range = utils.validate_and_clip_range(
                score_range, self._full_score_range, val_type=(int, float))
            filters.append(
                lambda record: (
                    record.get("score")
                    and score_range[0] <= record["score"] < score_range[1]
                )
            )
        return filters

    def _filtered_iter(self, filters):
        if filters:
            for record in self:
                if all(filter_(record) for filter_ in filters):
                    yield record
        else:
            for record in self:
                yield record

    def texts(
        self,
        *,
        subreddit: Optional[Union[str, Set[str]]] = None,
        date_range: Optional[Tuple[Optional[str], Optional[str]]] = None,
        score_range: Optional[Tuple[Optional[int], Optional[int]]] = None,
        min_len: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> Iterable[str]:
        """
        Iterate over comments (text-only) in 1 or more files of this dataset,
        optionally filtering by a variety of metadata and/or text length,
        in chronological order.

        Args:
            subreddit: Filter comments for those which were posted
                in the specified subreddit(s).
            date_range: Filter comments for those which were posted
                within the interval [start, end). Each item must be a str in
                ISO-standard format, i.e. some amount of YYYY-MM-DDTHH:mm:ss.
                Both start and end values must be specified, but a null value
                for either is automatically replaced by the minimum or maximum
                valid values, respectively.
            score_range: Filter comments for those whose score
                (# upvotes minus # downvotes) is within the interval [low, high).
                Both start and end values must be specified, but a null value
                for either is automatically replaced by the minimum or maximum
                valid values, respectively.
            min_len: Filter comments for those whose body length in chars is
                at least this long.
            limit: Maximum number of comments passing all filters to yield.
                If None, all comments are iterated over.

        Yields:
            Text of the next comment in dataset passing all filters.

        Raises:
            ValueError: If any filtering options are invalid.
        """
        self._date_range = date_range  # used to limit files iterated
        try:
            filters = self._get_filters(subreddit, date_range, score_range, min_len)
            for record in itertools.islice(self._filtered_iter(filters), limit):
                yield record["body"]
        finally:
            self._date_range = None

    def records(
        self,
        *,
        subreddit: Optional[Union[str, Set[str]]] = None,
        date_range: Optional[Tuple[Optional[str], Optional[str]]] = None,
        score_range: Optional[Tuple[Optional[int], Optional[int]]] = None,
        min_len: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> Iterable[Tuple[str, dict]]:
        """
        Iterate over comments (including text and metadata) in 1 or more files
        of this dataset, optionally filtering by a variety of metadata and/or
        text length, in chronological order.

        Args:
            subreddit: Filter comments for those which were posted
                in the specified subreddit(s).
            date_range: Filter comments for those which were posted
                within the interval [start, end). Each item must be a str in
                ISO-standard format, i.e. some amount of YYYY-MM-DDTHH:mm:ss.
                Both start and end values must be specified, but a null value
                for either is automatically replaced by the minimum or maximum
                valid values, respectively.
            score_range: Filter comments for those whose score
                (# upvotes minus # downvotes) is within the interval [low, high).
                Both start and end values must be specified, but a null value
                for either is automatically replaced by the minimum or maximum
                valid values, respectively.
            min_len: Filter comments for those whose body length in chars is
                at least this long.
            limit: Maximum number of comments passing all filters to yield.
                If None, all comments are iterated over.

        Yields:
            Text of the next comment in dataset passing all filters,
            and its corresponding metadata.

        Raises:
            ValueError: If any filtering options are invalid.
        """
        self._date_range = date_range  # used to limit files iterated
        try:
            filters = self._get_filters(subreddit, date_range, score_range, min_len)
            for record in itertools.islice(self._filtered_iter(filters), limit):
                yield record.pop("body"), record
        finally:
            self._date_range = None

    def _convert_timestamp(self, timestamp):
        try:
            return datetime.utcfromtimestamp(int(timestamp)).strftime(
                "%Y-%m-%dT%H:%M:%S"
            )
        except (ValueError, TypeError):
            return ""

    def _clean_content(self, content):
        # strip out link markup, e.g. [foo](http://foo.com)
        content = RE_REDDIT_LINK.sub(r"\1", content)
        # clean up basic HTML cruft
        content = content.replace("&gt;", ">").replace("&lt;", "<")
        # strip out text markup, e.g. * for bold text
        content = content.replace("`", "").replace("*", "").replace("~", "")
        # normalize whitespace
        return constants.RE_NONBREAKING_SPACE.sub(
            " ", constants.RE_LINEBREAK.sub(r"\n", content)
        ).strip()



"""
TextRank
--------
"""
import collections
import operator
from typing import cast, Callable, Collection, Dict, List, Optional, Set, Tuple, Union

from spacy.tokens import Doc, Token

from .. import utils
from . import graph_base
from . import utils as ke_utils


def textrank(
    doc: Doc,
    *,
    normalize: Optional[Union[str, Callable[[Token], str]]] = "lemma",
    include_pos: Optional[Union[str, Collection[str]]] = ("NOUN", "PROPN", "ADJ"),
    window_size: int = 2,
    edge_weighting: str = "binary",
    position_bias: bool = False,
    topn: Union[int, float] = 10,
) -> List[Tuple[str, float]]:
    """
    Extract key terms from a document using the TextRank algorithm, or
    a variation thereof. For example:

    - TextRank: ``window_size=2, edge_weighting="binary", position_bias=False``
    - SingleRank: ``window_size=10, edge_weighting="count", position_bias=False``
    - PositionRank: ``window_size=10, edge_weighting="count", position_bias=True``

    Args:
        doc: spaCy ``Doc`` from which to extract keyterms.
        normalize: If "lemma", lemmatize terms; if "lower", lowercase terms; if None,
            use the form of terms as they appeared in ``doc``; if a callable,
            must accept a ``Token`` and return a str,
            e.g. :func:`textacy.spacier.utils.get_normalized_text()`.
        include_pos: One or more POS tags with which to filter for good candidate keyterms.
            If None, include tokens of all POS tags
            (which also allows keyterm extraction from docs without POS-tagging.)
        window_size: Size of sliding window in which term co-occurrences are determined.
        edge_weighting ({"count", "binary"}): : If "count", the nodes for
            all co-occurring terms are connected by edges with weight equal to
            the number of times they co-occurred within a sliding window;
            if "binary", all such edges have weight = 1.
        position_bias: If True, bias the PageRank algorithm for weighting
            nodes in the word graph, such that words appearing earlier and more
            frequently in ``doc`` tend to get larger weights.
        topn: Number of top-ranked terms to return as key terms.
            If an integer, represents the absolute number; if a float, value
            must be in the interval (0.0, 1.0], which is converted to an int by
            ``int(round(len(set(candidates)) * topn))``.

    Returns:
        Sorted list of top ``topn`` key terms and their corresponding TextRank ranking scores.

    References:
        - Mihalcea, R., & Tarau, P. (2004, July). TextRank: Bringing order into texts.
          Association for Computational Linguistics.
        - Wan, Xiaojun and Jianguo Xiao. 2008. Single document keyphrase extraction
          using neighborhood knowledge. In Proceedings of the 23rd AAAI Conference
          on Artificial Intelligence, pages 855–860.
        - Florescu, C. and Cornelia, C. (2017). PositionRank: An Unsupervised Approach
          to Keyphrase Extraction from Scholarly Documents. In proceedings of ACL*,
          pages 1105-1115.
    """
    # validate / transform args
    include_pos = cast(Set[str], utils.to_collection(include_pos, str, set))
    if isinstance(topn, float):
        if not 0.0 < topn <= 1.0:
            raise ValueError(
                "topn={} is invalid; "
                "must be an int, or a float between 0.0 and 1.0".format(topn)
            )

    # bail out on empty docs
    if not doc:
        return []

    word_pos: Optional[Dict[str, float]]
    if position_bias is True:
        word_pos = collections.defaultdict(float)
        for word, norm_word in zip(doc, ke_utils.normalize_terms(doc, normalize)):
            word_pos[norm_word] += 1 / (word.i + 1)
        sum_word_pos = sum(word_pos.values())
        word_pos = {word: pos / sum_word_pos for word, pos in word_pos.items()}
    else:
        word_pos = None
    # build a graph from all words in doc, then score them
    graph = graph_base.build_graph_from_terms(
        [word for word in doc],
        normalize=normalize,
        window_size=window_size,
        edge_weighting=edge_weighting,
    )
    word_scores = graph_base.rank_nodes_by_pagerank(
        graph, weight="weight", personalization=word_pos)
    # generate a list of candidate terms
    candidates = _get_candidates(doc, normalize, include_pos)
    if isinstance(topn, float):
        topn = int(round(len(set(candidates)) * topn))
    # rank candidates by aggregating constituent word scores
    candidate_scores = {
        " ".join(candidate): sum(word_scores.get(word, 0.0) for word in candidate)
        for candidate in candidates
    }
    sorted_candidate_scores = sorted(
        candidate_scores.items(), key=operator.itemgetter(1, 0), reverse=True)
    return ke_utils.get_filtered_topn_terms(
        sorted_candidate_scores, topn, match_threshold=0.8)


def _get_candidates(
    doc: Doc,
    normalize: Optional[Union[str, Callable]],
    include_pos: Optional[Set[str]],
) -> Set[Tuple[str, ...]]:
    """
    Get a set of candidate terms to be scored by joining the longest
    subsequences of valid words -- non-stopword and non-punct, filtered to
    nouns, proper nouns, and adjectives if ``doc`` is POS-tagged -- then
    normalized into strings.
    """
    def _is_valid_tok(tok):
        return (
            not (tok.is_stop or tok.is_punct or tok.is_space)
            and (include_pos is None or tok.pos_ in include_pos)
        )

    candidates = ke_utils.get_longest_subsequence_candidates(doc, _is_valid_tok)
    return {
        tuple(ke_utils.normalize_terms(candidate, normalize))
        for candidate in candidates
    }



"""
Capitol Words Congressional speeches
------------------------------------

A collection of ~11k (almost all) speeches given by the main protagonists of the
2016 U.S. Presidential election that had previously served in the U.S. Congress --
including Hillary Clinton, Bernie Sanders, Barack Obama, Ted Cruz, and John Kasich --
from January 1996 through June 2016.

Records include the following data:

    - ``text``: Full text of the Congressperson's remarks.
    - ``title``: Title of the speech, in all caps.
    - ``date``: Date on which the speech was given, as an ISO-standard string.
    - ``speaker_name``: First and last name of the speaker.
    - ``speaker_party``: Political party of the speaker: "R" for Republican,
      "D" for Democrat, "I" for Independent.
    - ``congress``: Number of the Congress in which the speech was given: ranges
      continuously between 104 and 114.
    - ``chamber``: Chamber of Congress in which the speech was given: almost all
      are either "House" or "Senate", with a small number of "Extensions".

This dataset was derived from data provided by the (now defunct) Sunlight
Foundation's `Capitol Words API <http://sunlightlabs.github.io/Capitol-Words/>`_.
"""
import itertools
import logging
import pathlib
import urllib.parse
from typing import Iterable, Optional, Set, Tuple, Union

from .. import constants, utils
from .. import io as tio
from .base import Dataset

LOGGER = logging.getLogger(__name__)

NAME = "capitol_words"
META = {
    "site_url": "http://sunlightlabs.github.io/Capitol-Words/",
    "description": (
        "Collection of ~11k speeches in the Congressional Record given by "
        "notable U.S. politicians between Jan 1996 and Jun 2016."
    ),
}
DOWNLOAD_ROOT = "https://github.com/bdewilde/textacy-data/releases/download/"


class CapitolWords(Dataset):
    """
    Stream a collection of Congressional speeches from a compressed json file on disk,
    either as texts or text + metadata pairs.

    Download the data (one time only!) from the textacy-data repo
    (https://github.com/bdewilde/textacy-data), and save its contents to disk::

        >>> import textacy.datasets
        >>> ds = textacy.datasets.CapitolWords()
        >>> ds.download()
        >>> ds.info
        {'name': 'capitol_words',
         'site_url': 'http://sunlightlabs.github.io/Capitol-Words/',
         'description': 'Collection of ~11k speeches in the Congressional Record given by notable U.S. politicians between Jan 1996 and Jun 2016.'}

    Iterate over speeches as texts or records with both text and metadata::

        >>> for text in ds.texts(limit=3):
        ...     print(text, end="\\n\\n")
        >>> for text, meta in ds.records(limit=3):
        ...     print("\\n{} ({})\\n{}".format(meta["title"], meta["speaker_name"], text))

    Filter speeches by a variety of metadata fields and text length::

        >>> for text, meta in ds.records(speaker_name="Bernie Sanders", limit=3):
        ...     print("\\n{}, {}\\n{}".format(meta["title"], meta["date"], text))
        >>> for text, meta in ds.records(speaker_party="D", congress={110, 111, 112},
        ...                          chamber="Senate", limit=3):
        ...     print(meta["title"], meta["speaker_name"], meta["date"])
        >>> for text, meta in ds.records(speaker_name={"Barack Obama", "Hillary Clinton"},
        ...                              date_range=("2005-01-01", "2005-12-31")):
        ...     print(meta["title"], meta["speaker_name"], meta["date"])
        >>> for text in ds.texts(min_len=50000):
        ...     print(len(text))

    Stream speeches into a :class:`textacy.Corpus <textacy.corpus.Corpus>`::

        >>> textacy.Corpus("en", data=ota.records(limit=100))
        Corpus(100 docs; 70496 tokens)

    Args:
        data_dir: Path to directory on disk under which dataset is stored,
            i.e. ``/path/to/data_dir/capitol_words``.

    Attributes:
        full_date_range: First and last dates for which speeches are available,
            each as an ISO-formatted string (YYYY-MM-DD).
        speaker_names: Full names of all speakers included in corpus, e.g. "Bernie Sanders".
        speaker_parties: All distinct political parties of speakers, e.g. "R".
        chambers: All distinct chambers in which speeches were given, e.g. "House".
        congresses: All distinct numbers of the congresses in which speeches were given, e.g. 114.
    """

    full_date_range: Tuple[str, str] = ("1996-01-01", "2016-06-30")
    speaker_names: Set[str] = {
        "Barack Obama",
        "Bernie Sanders",
        "Hillary Clinton",
        "Jim Webb",
        "Joe Biden",
        "John Kasich",
        "Joseph Biden",
        "Lincoln Chafee",
        "Lindsey Graham",
        "Marco Rubio",
        "Mike Pence",
        "Rand Paul",
        "Rick Santorum",
        "Ted Cruz",
    }
    speaker_parties: Set[str] = {"D", "I", "R"}
    chambers: Set[str] = {"Extensions", "House", "Senate"}
    congresses: Set[int] = {104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114}

    def __init__(
        self,
        data_dir: Union[str, pathlib.Path] = constants.DEFAULT_DATA_DIR.joinpath(NAME),
    ):
        super().__init__(NAME, meta=META)
        self.data_dir = utils.to_path(data_dir).resolve()
        self._filename = "capitol-words-py3.json.gz"
        self._filepath = self.data_dir.joinpath(self._filename)

    @property
    def filepath(self) -> Optional[str]:
        """
        Full path on disk for CapitolWords data as compressed json file.
        ``None`` if file is not found, e.g. has not yet been downloaded.
        """
        if self._filepath.is_file():
            return str(self._filepath)
        else:
            return None

    def download(self, *, force: bool = False) -> None:
        """
        Download the data as a Python version-specific compressed json file and
        save it to disk under the ``data_dir`` directory.

        Args:
            force: If True, download the dataset, even if it already exists
                on disk under ``data_dir``.
        """
        release_tag = "capitol_words_py3_v{data_version}".format(data_version=1.0)
        url = urllib.parse.urljoin(DOWNLOAD_ROOT, release_tag + "/" + self._filename)
        tio.download_file(
            url,
            filename=self._filename,
            dirpath=self.data_dir,
            force=force,
        )

    def __iter__(self):
        if not self._filepath.is_file():
            raise OSError(
                "dataset file {} not found;\n"
                "has the dataset been downloaded yet?".format(self._filepath)
            )
        for record in tio.read_json(self._filepath, mode="rt", lines=True):
            yield record

    def _get_filters(
        self,
        speaker_name: Optional[Union[str, Set[str]]] = None,
        speaker_party: Optional[Union[str, Set[str]]] = None,
        chamber: Optional[Union[str, Set[str]]] = None,
        congress: Optional[Union[int, Set[int]]] = None,
        date_range: Optional[Tuple[Optional[str], Optional[str]]] = None,
        min_len: Optional[int] = None,
    ):
        filters = []
        if min_len is not None:
            if min_len < 1:
                raise ValueError("`min_len` must be at least 1")
            filters.append(
                lambda record: len(record.get("text", "")) >= min_len
            )
        if date_range is not None:
            date_range = utils.validate_and_clip_range(
                date_range, self.full_date_range, val_type=(str, bytes))
            filters.append(
                lambda record: (
                    record.get("date")
                    and date_range[0] <= record["date"] < date_range[1]
                )
            )
        if speaker_name is not None:
            speaker_name = utils.validate_set_members(
                speaker_name, (str, bytes), valid_vals=self.speaker_names)
            filters.append(lambda record: record.get("speaker_name") in speaker_name)
        if speaker_party is not None:
            speaker_party = utils.validate_set_members(
                speaker_party, (str, bytes), valid_vals=self.speaker_parties)
            filters.append(lambda record: record.get("speaker_party") in speaker_party)
        if chamber is not None:
            chamber = utils.validate_set_members(
                chamber, (str, bytes), valid_vals=self.chambers)
            filters.append(lambda record: record.get("chamber") in chamber)
        if congress is not None:
            congress = utils.validate_set_members(
                congress, int, valid_vals=self.congresses)
            filters.append(lambda record: record.get("congress") in congress)
        return filters

    def _filtered_iter(self, filters):
        if filters:
            for record in self:
                if all(filter_(record) for filter_ in filters):
                    yield record
        else:
            for record in self:
                yield record

    def texts(
        self,
        *,
        speaker_name: Optional[Union[str, Set[str]]] = None,
        speaker_party: Optional[Union[str, Set[str]]] = None,
        chamber: Optional[Union[str, Set[str]]] = None,
        congress: Optional[Union[int, Set[int]]] = None,
        date_range: Optional[Tuple[Optional[str], Optional[str]]] = None,
        min_len: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> Iterable[str]:
        """
        Iterate over speeches in this dataset, optionally filtering by a variety
        of metadata and/or text length, and yield texts only,
        in chronological order.

        Args:
            speaker_name: Filter speeches by the speakers' name;
                see :attr:`CapitolWords.speaker_names`.
            speaker_party: Filter speeches by the speakers' party;
                see :attr:`CapitolWords.speaker_parties`.
            chamber: Filter speeches by the chamber in which they were given;
                see :attr:`CapitolWords.chambers`.
            congress: Filter speeches by the congress in which they were given;
                see :attr:`CapitolWords.congresses`.
            date_range: Filter speeches by the date on which they were given.
                Both start and end date must be specified, but a null value for either
                will be replaced by the min/max date available for the dataset.
            min_len: Filter texts by the length (# characters) of their text content.
            limit: Yield no more than ``limit`` texts that match all specified filters.

        Yields:
            Full text of next (by chronological order) speech in dataset
            passing all filter params.

        Raises:
            ValueError: If any filtering options are invalid.
        """
        filters = self._get_filters(
            speaker_name, speaker_party, chamber, congress, date_range, min_len)
        for record in itertools.islice(self._filtered_iter(filters), limit):
            yield record["text"]

    def records(
        self,
        *,
        speaker_name: Optional[Union[str, Set[str]]] = None,
        speaker_party: Optional[Union[str, Set[str]]] = None,
        chamber: Optional[Union[str, Set[str]]] = None,
        congress: Optional[Union[int, Set[int]]] = None,
        date_range: Optional[Tuple[Optional[str], Optional[str]]] = None,
        min_len: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> Iterable[Tuple[str, dict]]:
        """
        Iterate over speeches in this dataset, optionally filtering by a variety
        of metadata and/or text length, and yield text + metadata pairs,
        in chronological order.

        Args:
            speaker_name: Filter speeches by the speakers' name;
                see :attr:`CapitolWords.speaker_names`.
            speaker_party: Filter speeches by the speakers' party;
                see :attr:`CapitolWords.speaker_parties`.
            chamber: Filter speeches by the chamber in which they were given;
                see :attr:`CapitolWords.chambers`.
            congress: Filter speeches by the congress in which they were given;
                see :attr:`CapitolWords.congresses`.
            date_range: Filter speeches by the date on which they were given.
                Both start and end date must be specified, but a null value for either
                will be replaced by the min/max date available for the dataset.
            min_len: Filter speeches by the length (# characters) of their text content.
            limit: Yield no more than ``limit`` speeches that match all specified filters.

        Yields:
            Full text of the next (by chronological order) speech in dataset
            passing all filters, and its corresponding metadata.

        Raises:
            ValueError: If any filtering options are invalid.
        """
        filters = self._get_filters(
            speaker_name, speaker_party, chamber, congress, date_range, min_len)
        for record in itertools.islice(self._filtered_iter(filters), limit):
            yield record.pop("text"), record



"""
Supreme Court decisions
-----------------------

A collection of ~8.4k (almost all) decisions issued by the U.S. Supreme Court
from November 1946 through June 2016 -- the "modern" era.

Records include the following data:

    - ``text``: Full text of the Court's decision.
    - ``case_name``: Name of the court case, in all caps.
    - ``argument_date``: Date on which the case was argued before the Court, as
      an ISO-formatted string ("YYYY-MM-DD").
    - ``decision_date``: Date on which the Court's decision was announced, as
      an ISO-formatted string ("YYYY-MM-DD").
    - ``decision_direction``: Ideological direction of the majority's decision:
      one of "conservative", "liberal", or "unspecifiable".
    - ``maj_opinion_author``: Name of the majority opinion's author, if available
      and identifiable, as an integer code whose mapping is given in
      :attr:`SupremeCourt.opinion_author_codes`.
    - ``n_maj_votes``: Number of justices voting in the majority.
    - ``n_min_votes``: Number of justices voting in the minority.
    - ``issue``: Subject matter of the case's core disagreement (e.g. "affirmative
      action") rather than its legal basis (e.g. "the equal protection clause"),
      as a string code whose mapping is given in :attr:`SupremeCourt.issue_codes`.
    - ``issue_area``: Higher-level categorization of the issue (e.g. "Civil Rights"),
      as an integer code whose mapping is given in :attr:`SupremeCourt.issue_area_codes`.
    - ``us_cite_id``: Citation identifier for each case according to the official
      United States Reports. Note: There are ~300 cases with duplicate ids,
      and it's not clear if that's "correct" or a data quality problem.

The text in this dataset was derived from FindLaw's searchable database of court
cases: http://caselaw.findlaw.com/court/us-supreme-court.

The metadata was extracted without modification from the Supreme Court Database:
Harold J. Spaeth, Lee Epstein, et al. 2016 Supreme Court Database, Version 2016
Release 1. http://supremecourtdatabase.org. Its license is CC BY-NC 3.0 US:
https://creativecommons.org/licenses/by-nc/3.0/us/.

This dataset's creation was inspired by a blog post by Emily Barry:
http://www.emilyinamillion.me/blog/2016/7/13/visualizing-supreme-court-topics-over-time.

The two datasets were merged through much munging and a carefully
trained model using the ``dedupe`` package. The model's duplicate threshold
was set so as to maximize the F-score where precision had twice as much
weight as recall. Still, given occasionally baffling inconsistencies in case
naming, citation ids, and decision dates, a very small percentage of texts
may be incorrectly matched to metadata. (Sorry.)
"""
import itertools
import logging
import pathlib
import urllib.parse
from typing import Dict, Iterable, Optional, Set, Tuple, Union

from .. import constants, utils
from .. import io as tio
from .base import Dataset

LOGGER = logging.getLogger(__name__)

NAME = "supreme_court"
META = {
    "site_url": "http://caselaw.findlaw.com/court/us-supreme-court",
    "description": (
        "Collection of ~8.4k decisions issued by the U.S. Supreme Court "
        "between November 1946 and June 2016."
    ),
}
DOWNLOAD_ROOT = "https://github.com/bdewilde/textacy-data/releases/download/"


class SupremeCourt(Dataset):
    """
    Stream a collection of U.S. Supreme Court decisions from a compressed json file on disk,
    either as texts or text + metadata pairs.

    Download the data (one time only!) from the textacy-data repo
    (https://github.com/bdewilde/textacy-data), and save its contents to disk::

        >>> import textacy.datasets
        >>> ds = textacy.datasets.SupremeCourt()
        >>> ds.download()
        >>> ds.info
        {'name': 'supreme_court',
         'site_url': 'http://caselaw.findlaw.com/court/us-supreme-court',
         'description': 'Collection of ~8.4k decisions issued by the U.S. Supreme Court between November 1946 and June 2016.'}

    Iterate over decisions as texts or records with both text and metadata::

        >>> for text in ds.texts(limit=3):
        ...     print(text[:500], end="\\n\\n")
        >>> for text, meta in ds.records(limit=3):
        ...     print("\\n{} ({})\\n{}".format(meta["case_name"], meta["decision_date"], text[:500]))

    Filter decisions by a variety of metadata fields and text length::

        >>> for text, meta in ds.records(opinion_author=109, limit=3):  # Notorious RBG!
        ...     print(meta["case_name"], meta["decision_direction"], meta["n_maj_votes"])
        >>> for text, meta in ds.records(decision_direction="liberal",
        ...                              issue_area={1, 9, 10}, limit=3):
        ...     print(meta["case_name"], meta["maj_opinion_author"], meta["n_maj_votes"])
        >>> for text, meta in ds.records(opinion_author=102, date_range=('1985-02-11', '1986-02-11')):
        ...     print("\\n{} ({})".format(meta["case_name"], meta["decision_date"]))
        ...     print(ds.issue_codes[meta["issue"]], "=>", meta["decision_direction"])
        >>> for text in ds.texts(min_len=250000):
        ...     print(len(text))

    Stream decisions into a :class:`textacy.Corpus <textacy.corpus.Corpus>`::

        >>> textacy.Corpus("en", data=ds.records(limit=25))
        Corpus(25 docs; 136696 tokens)

    Args:
        data_dir (str or :class:`pathlib.Path`): Path to directory on disk
            under which the data is stored, i.e. ``/path/to/data_dir/supreme_court``.

    Attributes:
        full_date_range: First and last dates for which decisions are available,
            each as an ISO-formatted string (YYYY-MM-DD).
        decision_directions: All distinct decision directions, e.g. "liberal".
        opinion_author_codes: Mapping of majority opinion authors,
            from id code to full name.
        issue_area_codes: Mapping of high-level issue area of the case's core disagreement,
            from id code to description.
        issue_codes: Mapping of the specific issue of the case's core disagreement,
            from id code to description.
    """

    full_date_range: Tuple[str, str] = ("1946-11-18", "2016-06-27")
    decision_directions: Set[str] = {"conservative", "liberal", "unspecifiable"}
    opinion_author_codes: Dict[int, Optional[str]] = {
        -1: None,
        1: "Jay, John",
        2: "Rutledge, John",
        3: "Cushing, William",
        4: "Wilson, James",
        5: "Blair,  John",
        6: "Iredell, James",
        7: "Johnson, Thomas",
        8: "Paterson, William",
        9: "Rutledge, John",
        10: "Chase, Samuel",
        11: "Ellsworth, Oliver",
        12: "Washington, Bushrod",
        13: "Moore, Alfred",
        14: "Marshall, John",
        15: "Johnson, William",
        16: "Livingston, Henry",
        17: "Todd, Thomas",
        18: "Duvall, Gabriel",
        19: "Story, Joseph",
        20: "Thompson, Smith",
        21: "Trimble, Robert",
        22: "McLean, John",
        23: "Baldwin, Henry",
        24: "Wayne, James",
        25: "Taney, Roger",
        26: "Barbour, Philip",
        27: "Catron, John",
        28: "McKinley, John",
        29: "Daniel, Peter",
        30: "Nelson, Samuel",
        31: "Woodbury, Levi",
        32: "Grier, Robert",
        33: "Curtis, Benjamin",
        34: "Campbell, John",
        35: "Clifford, Nathan",
        36: "Swayne, Noah",
        37: "Miller, Samuel",
        38: "Davis, David",
        39: "Field, Stephen",
        40: "Chase, Salmon",
        41: "Strong, William",
        42: "Bradley, Joseph",
        43: "Hunt, Ward",
        44: "Waite, Morrison",
        45: "Harlan, John",
        46: "Woods, William",
        47: "Matthews, Stanley",
        48: "Gray, Horace",
        49: "Blatchford, Samuel",
        50: "Lamar, Lucius",
        51: "Fuller, Melville",
        52: "Brewer, David",
        53: "Brown, Henry",
        54: "Shiras,  George",
        55: "Jackson, Howell",
        56: "White, Edward",
        57: "Peckham, Rufus",
        58: "McKenna, Joseph",
        59: "Holmes, Oliver",
        60: "Day, William",
        61: "Moody, William",
        62: "Lurton, Horace",
        63: "Hughes, Charles",
        64: "Van Devanter, Willis",
        65: "Lamar, Joseph",
        66: "Pitney, Mahlon",
        67: "McReynolds, James",
        68: "Brandeis, Louis",
        69: "Clarke, John",
        70: "Taft, William",
        71: "Sutherland, George",
        72: "Butler, Pierce",
        73: "Sanford, Edward",
        74: "Stone, Harlan",
        75: "Hughes, Charles",
        76: "Roberts, Owen",
        77: "Cardozo, Benjamin",
        78: "Black, Hugo",
        79: "Reed, Stanley",
        80: "Frankfurter, Felix",
        81: "Douglas, William",
        82: "Murphy, Francis",
        83: "Byrnes, James",
        84: "Jackson, Robert",
        85: "Rutledge, Wiley",
        86: "Burton, Harold",
        87: "Vinson, Fred",
        88: "Clark, Tom",
        89: "Minton, Sherman",
        90: "Warren, Earl",
        91: "Harlan, John",
        92: "Brennan, William",
        93: "Whittaker, Charles",
        94: "Stewart, Potter",
        95: "White, Byron",
        96: "Goldberg, Arthur",
        97: "Fortas, Abe",
        98: "Marshall, Thurgood",
        99: "Burger, Warren",
        100: "Blackmun, Harry",
        101: "Powell, Lewis",
        102: "Rehnquist, William",
        103: "Stevens, John",
        104: "O'Connor, Sandra",
        105: "Scalia, Antonin",
        106: "Kennedy, Anthony",
        107: "Souter, David",
        108: "Thomas, Clarence",
        109: "Ginsburg, Ruth",
        110: "Breyer, Stephen",
        111: "Roberts, John",
        112: "Alito, Samuel",
        113: "Sotomayor, Sonia",
        114: "Kagan, Elena",
    }
    issue_area_codes: Dict[int, Optional[str]] = {
        -1: None,
        1: "Criminal Procedure",
        2: "Civil Rights",
        3: "First Amendment",
        4: "Due Process",
        5: "Privacy",
        6: "Attorneys",
        7: "Unions",
        8: "Economic Activity",
        9: "Judicial Power",
        10: "Federalism",
        11: "Interstate Relations",
        12: "Federal Taxation",
        13: "Miscellaneous",
        14: "Private Action",
    }
    issue_codes: Dict[str, str] = {
        "100010": "federal-state ownership dispute (cf. Submerged Lands Act)",
        "100020": "federal pre-emption of state court jurisdiction",
        "100030": "federal pre-emption of state legislation or regulation. cf. state regulation of business. rarely involves union activity. Does not involve constitutional interpretation unless the Court says it does.",
        "100040": "Submerged Lands Act (cf. federal-state ownership dispute)",
        "100050": "national supremacy: commodities",
        "100060": "national supremacy: intergovernmental tax immunity",
        "100070": "national supremacy: marital and family relationships and property, including obligation of child support",
        "100080": "national supremacy: natural resources (cf. natural resources - environmental protection)",
        "100090": "national supremacy: pollution, air or water (cf. natural resources - environmental protection)",
        "10010": "involuntary confession",
        "100100": "national supremacy: public utilities (cf. federal public utilities regulation)",
        "100110": "national supremacy: state tax (cf. state tax)",
        "100120": "national supremacy: miscellaneous",
        "100130": "miscellaneous federalism",
        "10020": "habeas corpus",
        "10030": "plea bargaining: the constitutionality of and/or the circumstances of its exercise",
        "10040": "retroactivity (of newly announced or newly enacted constitutional or statutory rights)",
        "10050": "search and seizure (other than as pertains to vehicles or Crime Control Act)",
        "10060": "search and seizure, vehicles",
        "10070": "search and seizure, Crime Control Act",
        "10080": "contempt of court or congress",
        "10090": "self-incrimination (other than as pertains to Miranda or immunity from prosecution)",
        "10100": "Miranda warnings",
        "10110": "self-incrimination, immunity from prosecution",
        "10120": "right to counsel (cf. indigents appointment of counsel or inadequate representation)",
        "10130": "cruel and unusual punishment, death penalty (cf. extra legal jury influence, death penalty)",
        "10140": "cruel and unusual punishment, non-death penalty (cf. liability, civil rights acts)",
        "10150": "line-up",
        "10160": "discovery and inspection (in the context of criminal litigation only, otherwise Freedom of Information Act and related federal or state statutes or regulations)",
        "10170": "double jeopardy",
        "10180": "ex post facto (state)",
        "10190": "extra-legal jury influences: miscellaneous",
        "10200": "extra-legal jury influences: prejudicial statements or evidence",
        "10210": "extra-legal jury influences: contact with jurors outside courtroom",
        "10220": "extra-legal jury influences: jury instructions (not necessarily in criminal cases)",
        "10230": "extra-legal jury influences: voir dire (not necessarily a criminal case)",
        "10240": "extra-legal jury influences: prison garb or appearance",
        "10250": "extra-legal jury influences: jurors and death penalty (cf. cruel and unusual punishment)",
        "10260": "extra-legal jury influences: pretrial publicity",
        "10270": "confrontation (right to confront accuser, call and cross-examine witnesses)",
        "10280": "subconstitutional fair procedure: confession of error",
        "10290": "subconstitutional fair procedure: conspiracy (cf. Federal Rules of Criminal Procedure: conspiracy)",
        "10300": "subconstitutional fair procedure: entrapment",
        "10310": "subconstitutional fair procedure: exhaustion of remedies",
        "10320": "subconstitutional fair procedure: fugitive from justice",
        "10330": "subconstitutional fair procedure: presentation, admissibility, or sufficiency of evidence (not necessarily a criminal case)",
        "10340": "subconstitutional fair procedure: stay of execution",
        "10350": "subconstitutional fair procedure: timeliness",
        "10360": "subconstitutional fair procedure: miscellaneous",
        "10370": "Federal Rules of Criminal Procedure",
        "10380": "statutory construction of criminal laws: assault",
        "10390": "statutory construction of criminal laws: bank robbery",
        "10400": "statutory construction of criminal laws: conspiracy (cf. subconstitutional fair procedure: conspiracy)",
        "10410": "statutory construction of criminal laws: escape from custody",
        "10420": "statutory construction of criminal laws: false statements (cf. statutory construction of criminal laws: perjury)",
        "10430": "statutory construction of criminal laws: financial (other than in fraud or internal revenue)",
        "10440": "statutory construction of criminal laws: firearms",
        "10450": "statutory construction of criminal laws: fraud",
        "10460": "statutory construction of criminal laws: gambling",
        "10470": "statutory construction of criminal laws: Hobbs Act; i.e., 18 USC 1951",
        "10480": "statutory construction of criminal laws: immigration (cf. immigration and naturalization)",
        "10490": "statutory construction of criminal laws: internal revenue (cf. Federal Taxation)",
        "10500": "statutory construction of criminal laws: Mann Act and related statutes",
        "10510": "statutory construction of criminal laws: narcotics includes regulation and prohibition of alcohol",
        "10520": "statutory construction of criminal laws: obstruction of justice",
        "10530": "statutory construction of criminal laws: perjury (other than as pertains to statutory construction of criminal laws: false statements)",
        "10540": "statutory construction of criminal laws: Travel Act, 18 USC 1952",
        "10550": "statutory construction of criminal laws: war crimes",
        "10560": "statutory construction of criminal laws: sentencing guidelines",
        "10570": "statutory construction of criminal laws: miscellaneous",
        "10580": "jury trial (right to, as distinct from extra-legal jury influences)",
        "10590": "speedy trial",
        "10600": "miscellaneous criminal procedure (cf. due process, prisoners' rights, comity: criminal procedure)",
        "110010": "boundary dispute between states",
        "110020": "non-real property dispute between states",
        "110030": "miscellaneous interstate relations conflict",
        "110033": "incorporation of foreign territories",
        "120010": "federal taxation, typically under provisions of the Internal Revenue Code",
        "120020": "federal taxation of gifts, personal, business, or professional expenses",
        "120030": "priority of federal fiscal claims: over those of the states or private entities",
        "120040": "miscellaneous federal taxation (cf. national supremacy: state tax)",
        "130010": "legislative veto",
        "130015": "executive authority vis-a-vis congress or the states",
        "130020": "miscellaneous",
        "140010": "real property",
        "140020": "personal property",
        "140030": "contracts",
        "140040": "evidence",
        "140050": "civil procedure",
        "140060": "torts",
        "140070": "wills and trusts",
        "140080": "commercial transactions",
        "20010": "voting",
        "20020": "Voting Rights Act of 1965, plus amendments",
        "20030": "ballot access (of candidates and political parties)",
        "20040": "desegregation (other than as pertains to school desegregation, employment discrimination, and affirmative action)",
        "20050": "desegregation, schools",
        "20060": "employment discrimination: on basis of race, age, religion, illegitimacy, national origin, or working conditions.",
        "20070": "affirmative action",
        "20075": "slavery or indenture",
        "20080": "sit-in demonstrations (protests against racial discrimination in places of public accommodation)",
        "20090": "reapportionment: other than plans governed by the Voting Rights Act",
        "20100": "debtors' rights",
        "20110": "deportation (cf. immigration and naturalization)",
        "20120": "employability of aliens (cf. immigration and naturalization)",
        "20130": "sex discrimination (excluding sex discrimination in employment)",
        "20140": "sex discrimination in employment (cf. sex discrimination)",
        "20150": "Indians (other than pertains to state jurisdiction over)",
        "20160": "Indians, state jurisdiction over",
        "20170": "juveniles (cf. rights of illegitimates)",
        "20180": "poverty law, constitutional",
        "20190": "poverty law, statutory: welfare benefits, typically under some Social Security Act provision.",
        "20200": "illegitimates, rights of (cf. juveniles): typically inheritance and survivor's benefits, and paternity suits",
        "20210": "handicapped, rights of: under Rehabilitation, Americans with Disabilities Act, and related statutes",
        "20220": "residency requirements: durational, plus discrimination against nonresidents",
        "20230": "military: draftee, or person subject to induction",
        "20240": "military: active duty",
        "20250": "military: veteran",
        "20260": "immigration and naturalization: permanent residence",
        "20270": "immigration and naturalization: citizenship",
        "20280": "immigration and naturalization: loss of citizenship, denaturalization",
        "20290": "immigration and naturalization: access to public education",
        "20300": "immigration and naturalization: welfare benefits",
        "20310": "immigration and naturalization: miscellaneous",
        "20320": "indigents: appointment of counsel (cf. right to counsel)",
        "20330": "indigents: inadequate representation by counsel (cf. right to counsel)",
        "20340": "indigents: payment of fine",
        "20350": "indigents: costs or filing fees",
        "20360": "indigents: U.S. Supreme Court docketing fee",
        "20370": "indigents: transcript",
        "20380": "indigents: assistance of psychiatrist",
        "20390": "indigents: miscellaneous",
        "20400": "liability, civil rights acts (cf. liability, governmental and liability, nongovernmental; cruel and unusual punishment, non-death penalty)",
        "20410": "miscellaneous civil rights (cf. comity: civil rights)",
        "30010": "First Amendment, miscellaneous (cf. comity: First Amendment)",
        "30020": "commercial speech, excluding attorneys",
        "30030": "libel, defamation: defamation of public officials and public and private persons",
        "30040": "libel, privacy: true and false light invasions of privacy",
        "30050": "legislative investigations: concerning internal security only",
        "30060": "federal or state internal security legislation: Smith, Internal Security, and related federal statutes",
        "30070": "loyalty oath or non-Communist affidavit (other than bar applicants, government employees, political party, or teacher)",
        "30080": "loyalty oath: bar applicants (cf. admission to bar, state or federal or U.S. Supreme Court)",
        "30090": "loyalty oath: government employees",
        "30100": "loyalty oath: political party",
        "30110": "loyalty oath: teachers",
        "30120": "security risks: denial of benefits or dismissal of employees for reasons other than failure to meet loyalty oath requirements",
        "30130": "conscientious objectors (cf. military draftee or military active duty) to military service",
        "30140": "campaign spending (cf. governmental corruption):",
        "30150": "protest demonstrations (other than as pertains to sit-in demonstrations): demonstrations and other forms of protest based on First Amendment guarantees",
        "30160": "free exercise of religion",
        "30170": "establishment of religion (other than as pertains to parochiaid:)",
        "30180": "parochiaid: government aid to religious schools, or religious requirements in public schools",
        "30190": "obscenity, state (cf. comity: privacy): including the regulation of sexually explicit material under the 21st Amendment",
        "30200": "obscenity, federal",
        "40010": "due process: miscellaneous (cf. loyalty oath), the residual code",
        "40020": "due process: hearing or notice (other than as pertains to government employees or prisoners' rights)",
        "40030": "due process: hearing, government employees",
        "40040": "due process: prisoners' rights and defendants' rights",
        "40050": "due process: impartial decision maker",
        "40060": "due process: jurisdiction (jurisdiction over non-resident litigants)",
        "40070": "due process: takings clause, or other non-constitutional governmental taking of property",
        "50010": "privacy (cf. libel, comity: privacy)",
        "50020": "abortion: including contraceptives",
        "50030": "right to die",
        "50040": "Freedom of Information Act and related federal or state statutes or regulations",
        "60010": "attorneys' and governmental employees' or officials' fees or compensation or licenses",
        "60020": "commercial speech, attorneys (cf. commercial speech)",
        "60030": "admission to a state or federal bar, disbarment, and attorney discipline (cf. loyalty oath: bar applicants)",
        "60040": "admission to, or disbarment from, Bar of the U.S. Supreme Court",
        "70010": "arbitration (in the context of labor-management or employer-employee relations) (cf. arbitration)",
        "70020": "union antitrust: legality of anticompetitive union activity",
        "70030": "union or closed shop: includes agency shop litigation",
        "70040": "Fair Labor Standards Act",
        "70050": "Occupational Safety and Health Act",
        "70060": "union-union member dispute (except as pertains to union or closed shop)",
        "70070": "labor-management disputes: bargaining",
        "70080": "labor-management disputes: employee discharge",
        "70090": "labor-management disputes: distribution of union literature",
        "70100": "labor-management disputes: representative election",
        "70110": "labor-management disputes: antistrike injunction",
        "70120": "labor-management disputes: jurisdictional dispute",
        "70130": "labor-management disputes: right to organize",
        "70140": "labor-management disputes: picketing",
        "70150": "labor-management disputes: secondary activity",
        "70160": "labor-management disputes: no-strike clause",
        "70170": "labor-management disputes: union representatives",
        "70180": "labor-management disputes: union trust funds (cf. ERISA)",
        "70190": "labor-management disputes: working conditions",
        "70200": "labor-management disputes: miscellaneous dispute",
        "70210": "miscellaneous union",
        "80010": "antitrust (except in the context of mergers and union antitrust)",
        "80020": "mergers",
        "80030": "bankruptcy (except in the context of priority of federal fiscal claims)",
        "80040": "sufficiency of evidence: typically in the context of a jury's determination of compensation for injury or death",
        "80050": "election of remedies: legal remedies available to injured persons or things",
        "80060": "liability, governmental: tort or contract actions by or against government or governmental officials other than defense of criminal actions brought under a civil rights action.",
        "80070": "liability, other than as in sufficiency of evidence, election of remedies, punitive damages",
        "80080": "liability, punitive damages",
        "80090": "Employee Retirement Income Security Act (cf. union trust funds)",
        "80100": "state or local government tax",
        "80105": "state and territorial land claims",
        "80110": "state or local government regulation, especially of business (cf. federal pre-emption of state court jurisdiction, federal pre-emption of state legislation or regulation)",
        "80120": "federal or state regulation of securities",
        "80130": "natural resources - environmental protection (cf. national supremacy: natural resources, national supremacy: pollution)",
        "80140": "corruption, governmental or governmental regulation of other than as in campaign spending",
        "80150": "zoning: constitutionality of such ordinances, or restrictions on owners' or lessors' use of real property",
        "80160": "arbitration (other than as pertains to labor-management or employer-employee relations (cf. union arbitration)",
        "80170": "federal or state consumer protection: typically under the Truth in Lending; Food, Drug and Cosmetic; and Consumer Protection Credit Acts",
        "80180": "patents and copyrights: patent",
        "80190": "patents and copyrights: copyright",
        "80200": "patents and copyrights: trademark",
        "80210": "patents and copyrights: patentability of computer processes",
        "80220": "federal or state regulation of transportation regulation: railroad",
        "80230": "federal and some few state regulations of transportation regulation: boat",
        "80240": "federal and some few state regulation of transportation regulation:truck, or motor carrier",
        "80250": "federal and some few state regulation of transportation regulation: pipeline (cf. federal public utilities regulation: gas pipeline)",
        "80260": "federal and some few state regulation of transportation regulation: airline",
        "80270": "federal and some few state regulation of public utilities regulation: electric power",
        "80280": "federal and some few state regulation of public utilities regulation: nuclear power",
        "80290": "federal and some few state regulation of public utilities regulation: oil producer",
        "80300": "federal and some few state regulation of public utilities regulation: gas producer",
        "80310": "federal and some few state regulation of public utilities regulation: gas pipeline (cf. federal transportation regulation: pipeline)",
        "80320": "federal and some few state regulation of public utilities regulation: radio and television (cf. cable television)",
        "80330": "federal and some few state regulation of public utilities regulation: cable television (cf. radio and television)",
        "80340": "federal and some few state regulations of public utilities regulation: telephone or telegraph company",
        "80350": "miscellaneous economic regulation",
        "90010": "comity: civil rights",
        "90020": "comity: criminal procedure",
        "90030": "comity: First Amendment",
        "90040": "comity: habeas corpus",
        "90050": "comity: military",
        "90060": "comity: obscenity",
        "90070": "comity: privacy",
        "90080": "comity: miscellaneous",
        "90090": "comity primarily removal cases, civil procedure (cf. comity, criminal and First Amendment); deference to foreign judicial tribunals",
        "90100": "assessment of costs or damages: as part of a court order",
        "90110": "Federal Rules of Civil Procedure including Supreme Court Rules, application of the Federal Rules of Evidence, Federal Rules of Appellate Procedure in civil litigation, Circuit Court Rules, and state rules and admiralty rules",
        "90120": "judicial review of administrative agency's or administrative official's actions and procedures",
        "90130": "mootness (cf. standing to sue: live dispute)",
        "90140": "venue",
        "90150": "no merits: writ improvidently granted",
        "90160": "no merits: dismissed or affirmed for want of a substantial or properly presented federal question, or a nonsuit",
        "90170": "no merits: dismissed or affirmed for want of jurisdiction (cf. judicial administration: Supreme Court jurisdiction or authority on appeal from federal district courts or courts of appeals)",
        "90180": "no merits: adequate non-federal grounds for decision",
        "90190": "no merits: remand to determine basis of state or federal court decision (cf. judicial administration: state law)",
        "90200": "no merits: miscellaneous",
        "90210": "standing to sue: adversary parties",
        "90220": "standing to sue: direct injury",
        "90230": "standing to sue: legal injury",
        "90240": "standing to sue: personal injury",
        "90250": "standing to sue: justiciable question",
        "90260": "standing to sue: live dispute",
        "90270": "standing to sue: parens patriae standing",
        "90280": "standing to sue: statutory standing",
        "90290": "standing to sue: private or implied cause of action",
        "90300": "standing to sue: taxpayer's suit",
        "90310": "standing to sue: miscellaneous",
        "90320": "judicial administration: jurisdiction or authority of federal district courts or territorial courts",
        "90330": "judicial administration: jurisdiction or authority of federal courts of appeals",
        "90340": "judicial administration: Supreme Court jurisdiction or authority on appeal or writ of error, from federal district courts or courts of appeals (cf. 753)",
        "90350": "judicial administration: Supreme Court jurisdiction or authority on appeal or writ of error, from highest state court",
        "90360": "judicial administration: jurisdiction or authority of the Court of Claims",
        "90370": "judicial administration: Supreme Court's original jurisdiction",
        "90380": "judicial administration: review of non-final order",
        "90390": "judicial administration: change in state law (cf. no merits: remand to determine basis of state court decision)",
        "90400": "judicial administration: federal question (cf. no merits: dismissed for want of a substantial or properly presented federal question)",
        "90410": "judicial administration: ancillary or pendent jurisdiction",
        "90420": "judicial administration: extraordinary relief (e.g., mandamus, injunction)",
        "90430": "judicial administration: certification (cf. objection to reason for denial of certiorari or appeal)",
        "90440": "judicial administration: resolution of circuit conflict, or conflict between or among other courts",
        "90450": "judicial administration: objection to reason for denial of certiorari or appeal",
        "90460": "judicial administration: collateral estoppel or res judicata",
        "90470": "judicial administration: interpleader",
        "90480": "judicial administration: untimely filing",
        "90490": "judicial administration: Act of State doctrine",
        "90500": "judicial administration: miscellaneous",
        "90510": "Supreme Court's certiorari, writ of error, or appeals jurisdiction",
        "90520": "miscellaneous judicial power, especially diversity jurisdiction",
    }

    def __init__(
        self,
        data_dir: Union[str, pathlib.Path] = constants.DEFAULT_DATA_DIR.joinpath(NAME),
    ):
        super().__init__(NAME, meta=META)
        self.data_dir = utils.to_path(data_dir).resolve()
        self._filename = "supreme-court-py3.json.gz"
        self._filepath = self.data_dir.joinpath(self._filename)

    @property
    def filepath(self) -> Optional[str]:
        """
        Full path on disk for SupremeCourt data as compressed json file.
        ``None`` if file is not found, e.g. has not yet been downloaded.
        """
        if self._filepath.is_file():
            return str(self._filepath)
        else:
            return None

    def download(self, *, force: bool = False) -> None:
        """
        Download the data as a Python version-specific compressed json file and
        save it to disk under the ``data_dir`` directory.

        Args:
            force: If True, download the dataset, even if it already exists
                on disk under ``data_dir``.
        """
        release_tag = "supreme_court_py3_v{data_version}".format(data_version=1.0)
        url = urllib.parse.urljoin(DOWNLOAD_ROOT, release_tag + "/" + self._filename)
        tio.download_file(
            url,
            filename=self._filename,
            dirpath=self.data_dir,
            force=force,
        )

    def __iter__(self):
        if not self._filepath.is_file():
            raise OSError(
                "dataset file {} not found;\n"
                "has the dataset been downloaded yet?".format(self._filepath)
            )
        for record in tio.read_json(self._filepath, mode="rt", lines=True):
            yield record

    def _get_filters(
        self,
        opinion_author,
        decision_direction,
        issue_area,
        date_range,
        min_len,
    ):
        filters = []
        if min_len is not None:
            if min_len < 1:
                raise ValueError("`min_len` must be at least 1")
            filters.append(
                lambda record: len(record.get("text", "")) >= min_len
            )
        if date_range is not None:
            date_range = utils.validate_and_clip_range(
                date_range, self.full_date_range, val_type=(str, bytes))
            filters.append(
                lambda record: (
                    record.get("decision_date")
                    and date_range[0] <= record["decision_date"] < date_range[1]
                )
            )
        if opinion_author is not None:
            opinion_author = utils.validate_set_members(
                opinion_author, int, valid_vals=self.opinion_author_codes)
            filters.append(
                lambda record: record.get("maj_opinion_author") in opinion_author)
        if decision_direction is not None:
            decision_direction = utils.validate_set_members(
                decision_direction, (str, bytes), valid_vals=self.decision_directions)
            filters.append(
                lambda record: record.get("decision_direction") in decision_direction)
        if issue_area is not None:
            issue_area = utils.validate_set_members(
                issue_area, int, valid_vals=self.issue_area_codes)
            filters.append(
                lambda record: record.get("issue_area") in issue_area)
        return filters

    def _filtered_iter(self, filters):
        if filters:
            for record in self:
                if all(filter_(record) for filter_ in filters):
                    yield record
        else:
            for record in self:
                yield record

    def texts(
        self,
        *,
        opinion_author: Optional[Union[int, Set[int]]] = None,
        decision_direction: Optional[Union[str, Set[str]]] =None,
        issue_area: Optional[Union[int, Set[int]]] = None,
        date_range: Optional[Tuple[Optional[str], Optional[str]]] = None,
        min_len: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> Iterable[str]:
        """
        Iterate over decisions in this dataset, optionally filtering by a variety
        of metadata and/or text length, and yield texts only,
        in chronological order by decision date.

        Args:
            opinion_author: Filter decisions by the name(s) of the majority opinion's author,
                coded as an integer whose mapping is given in
                :attr:`SupremeCourt.opinion_author_codes`.
            decision_direction: Filter decisions by the ideological direction
                of the majority's decision; see
                :attr:`SupremeCourt.decision_directions`.
            issue_area: Filter decisions by the issue area of the case's subject matter,
                coded as an integer whose mapping is given in
                :attr:`SupremeCourt.issue_area_codes`.
            date_range: Filter decisions by the date on which they were decided;
                both start and end date must be specified, but a null value for either
                will be replaced by the min/max date available for the dataset.
            min_len: Filter decisions by the length (# characters) of their text content.
            limit: Yield no more than ``limit`` decisions that match all specified filters.

        Yields:
            Text of the next decision in dataset passing all filters.

        Raises:
            ValueError: If any filtering options are invalid.
        """
        filters = self._get_filters(
            opinion_author, decision_direction, issue_area, date_range, min_len)
        for record in itertools.islice(self._filtered_iter(filters), limit):
            yield record["text"]

    def records(
        self,
        *,
        opinion_author: Optional[Union[int, Set[int]]] = None,
        decision_direction: Optional[Union[str, Set[str]]] =None,
        issue_area: Optional[Union[int, Set[int]]] = None,
        date_range: Optional[Tuple[Optional[str], Optional[str]]] = None,
        min_len: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> Iterable[Tuple[str, dict]]:
        """
        Iterate over decisions in this dataset, optionally filtering by a variety
        of metadata and/or text length, and yield text + metadata pairs,
        in chronological order by decision date.

        Args:
            opinion_author: Filter decisions by the name(s) of the majority opinion's author,
                coded as an integer whose mapping is given in
                :attr:`SupremeCourt.opinion_author_codes`.
            decision_direction: Filter decisions by the ideological direction
                of the majority's decision; see
                :attr:`SupremeCourt.decision_directions`.
            issue_area: Filter decisions by the issue area of the case's subject matter,
                coded as an integer whose mapping is given in
                :attr:`SupremeCourt.issue_area_codes`.
            date_range: Filter decisions by the date on which they were decided;
                both start and end date must be specified, but a null value for either
                will be replaced by the min/max date available for the dataset.
            min_len: Filter decisions by the length (# characters) of their text content.
            limit: Yield no more than ``limit`` decisions that match all specified filters.

        Yields:
            Text of the next decision in dataset passing all filters,
            and its corresponding metadata.

        Raises:
            ValueError: If any filtering options are invalid.
        """
        filters = self._get_filters(
            opinion_author, decision_direction, issue_area, date_range, min_len)
        for record in itertools.islice(self._filtered_iter(filters), limit):
            yield record.pop("text"), record



"""
Functions to extract various elements of interest from documents already parsed
by spaCy, such as n-grams, named entities, subject-verb-object triples, and
acronyms.
"""
import collections
import itertools
import operator
import re
from typing import cast, Callable, Dict, Iterable, List, Optional, Set, Tuple, Union

import numpy as np
from cytoolz import itertoolz
from spacy.parts_of_speech import CONJ, DET, NOUN, VERB
from spacy.matcher import Matcher
from spacy.tokens import Doc, Span, Token

from . import constants
from . import text_utils
from . import utils
from .spacier import utils as spacy_utils


def words(
    doc: Union[Doc, Span],
    *,
    filter_stops: bool = True,
    filter_punct: bool = True,
    filter_nums: bool = False,
    include_pos: Optional[Union[str, Set[str]]] = None,
    exclude_pos: Optional[Union[str, Set[str]]] = None,
    min_freq: int = 1,
) -> Iterable[Token]:
    """
    Extract an ordered sequence of words from a document processed by spaCy,
    optionally filtering words by part-of-speech tag and frequency.

    Args:
        doc
        filter_stops: If True, remove stop words from word list.
        filter_punct: If True, remove punctuation from word list.
        filter_nums: If True, remove number-like words (e.g. 10, "ten")
            from word list.
        include_pos: Remove words whose part-of-speech tag IS NOT in the specified tags.
        exclude_pos: Remove words whose part-of-speech tag IS in the specified tags.
        min_freq: Remove words that occur in ``doc`` fewer than ``min_freq`` times.

    Yields:
        Next token from ``doc`` passing specified filters in order of appearance
        in the document.

    Raises:
        TypeError: if ``include_pos`` or ``exclude_pos`` is not a str, a set of str,
            or a falsy value

    Note:
        Filtering by part-of-speech tag uses the universal POS tag set; for details,
        check spaCy's docs: https://spacy.io/api/annotation#pos-tagging
    """
    words_ = (w for w in doc if not w.is_space)
    if filter_stops is True:
        words_ = (w for w in words_ if not w.is_stop)
    if filter_punct is True:
        words_ = (w for w in words_ if not w.is_punct)
    if filter_nums is True:
        words_ = (w for w in words_ if not w.like_num)
    if include_pos:
        include_pos = cast(Set[str], utils.to_collection(include_pos, str, set))
        include_pos = {pos.upper() for pos in include_pos}
        words_ = (w for w in words_ if w.pos_ in include_pos)
    if exclude_pos:
        exclude_pos = cast(Set[str], utils.to_collection(exclude_pos, str, set))
        exclude_pos = {pos.upper() for pos in exclude_pos}
        words_ = (w for w in words_ if w.pos_ not in exclude_pos)
    if min_freq > 1:
        words_ = list(words_)
        freqs = itertoolz.frequencies(w.lower_ for w in words_)
        words_ = (w for w in words_ if freqs[w.lower_] >= min_freq)

    for word in words_:
        yield word


def ngrams(
    doc: Union[Doc, Span],
    n: int,
    *,
    filter_stops: bool = True,
    filter_punct: bool = True,
    filter_nums: bool = False,
    include_pos: Optional[Union[str, Set[str]]] = None,
    exclude_pos: Optional[Union[str, Set[str]]] = None,
    min_freq: int = 1,
) -> Iterable[Span]:
    """
    Extract an ordered sequence of n-grams (``n`` consecutive words) from a
    spacy-parsed doc, optionally filtering n-grams by the types and
    parts-of-speech of the constituent words.

    Args:
        doc
        n: Number of tokens per n-gram; 2 => bigrams, 3 => trigrams, etc.
        filter_stops: If True, remove ngrams that start or end with a stop word
        filter_punct: If True, remove ngrams that contain any punctuation-only tokens
        filter_nums: If True, remove ngrams that contain any numbers
            or number-like tokens (e.g. 10, 'ten')
        include_pos: Remove ngrams if any of their constituent tokens' part-of-speech tags
            ARE NOT included in this param
        exclude_pos: Remove ngrams if any of their constituent tokens' part-of-speech tags
            ARE included in this param
        min_freq: Remove ngrams that occur in ``doc`` fewer than ``min_freq`` times

    Yields:
        Next ngram from ``doc`` passing all specified filters, in order of appearance
        in the document

    Raises:
        ValueError: if ``n`` < 1
        TypeError: if ``include_pos`` or ``exclude_pos`` is not a str, a set of str,
            or a falsy value

    Note:
        Filtering by part-of-speech tag uses the universal POS tag set; for details,
        check spaCy's docs: https://spacy.io/api/annotation#pos-tagging
    """
    if n < 1:
        raise ValueError("n must be greater than or equal to 1")

    ngrams_ = (doc[i : i + n] for i in range(len(doc) - n + 1))
    ngrams_ = (ngram for ngram in ngrams_ if not any(w.is_space for w in ngram))
    if filter_stops is True:
        ngrams_ = (
            ngram for ngram in ngrams_ if not ngram[0].is_stop and not ngram[-1].is_stop
        )
    if filter_punct is True:
        ngrams_ = (ngram for ngram in ngrams_ if not any(w.is_punct for w in ngram))
    if filter_nums is True:
        ngrams_ = (ngram for ngram in ngrams_ if not any(w.like_num for w in ngram))
    if include_pos:
        include_pos = cast(Set[str], utils.to_collection(include_pos, str, set))
        include_pos = {pos.upper() for pos in include_pos}
        ngrams_ = (
            ngram for ngram in ngrams_ if all(w.pos_ in include_pos for w in ngram)
        )
    if exclude_pos:
        exclude_pos = cast(Set[str], utils.to_collection(exclude_pos, str, set))
        exclude_pos = {pos.upper() for pos in exclude_pos}
        ngrams_ = (
            ngram
            for ngram in ngrams_
            if not any(w.pos_ in exclude_pos for w in ngram)
        )
    if min_freq > 1:
        ngrams_ = list(ngrams_)
        freqs = itertoolz.frequencies(ngram.lower_ for ngram in ngrams_)
        ngrams_ = (ngram for ngram in ngrams_ if freqs[ngram.lower_] >= min_freq)

    for ngram in ngrams_:
        yield ngram


def entities(
    doc: Doc,
    *,
    include_types: Optional[Union[str, Set[str]]] = None,
    exclude_types: Optional[Union[str, Set[str]]] = None,
    drop_determiners: bool = True,
    min_freq: int = 1,
) -> Iterable[Span]:
    """
    Extract an ordered sequence of named entities (PERSON, ORG, LOC, etc.) from
    a ``Doc``, optionally filtering by entity types and frequencies.

    Args:
        doc
        include_types: Remove entities whose type IS NOT
            in this param; if "NUMERIC", all numeric entity types ("DATE",
            "MONEY", "ORDINAL", etc.) are included
        exclude_types: Remove entities whose type IS
            in this param; if "NUMERIC", all numeric entity types ("DATE",
            "MONEY", "ORDINAL", etc.) are excluded
        drop_determiners: Remove leading determiners (e.g. "the")
            from entities (e.g. "the United States" => "United States").

            .. note:: Entities from which a leading determiner has been removed
               are, effectively, *new* entities, and not saved to the ``Doc``
               from which they came. This is irritating but unavoidable, since
               this function is not meant to have side-effects on document state.
               If you're only using the text of the returned spans, this is no
               big deal, but watch out if you're counting on determiner-less
               entities associated with the doc downstream.

        min_freq: Remove entities that occur in ``doc`` fewer
            than ``min_freq`` times

    Yields:
        Next entity from ``doc`` passing all specified filters in order of appearance
        in the document

    Raises:
        TypeError: if ``include_types`` or ``exclude_types`` is not a str, a set of
            str, or a falsy value
    """
    ents = doc.ents
    # HACK: spacy's models have been erroneously tagging whitespace as entities
    # https://github.com/explosion/spaCy/commit/1e6725e9b734862e61081a916baf440697b9971e
    ents = (ent for ent in ents if not ent.text.isspace())
    include_types = _parse_ent_types(include_types, "include")
    exclude_types = _parse_ent_types(exclude_types, "exclude")
    if include_types:
        if isinstance(include_types, str):
            ents = (ent for ent in ents if ent.label_ == include_types)
        elif isinstance(include_types, (set, frozenset, list, tuple)):
            ents = (ent for ent in ents if ent.label_ in include_types)
    if exclude_types:
        if isinstance(exclude_types, str):
            ents = (ent for ent in ents if ent.label_ != exclude_types)
        elif isinstance(exclude_types, (set, frozenset, list, tuple)):
            ents = (ent for ent in ents if ent.label_ not in exclude_types)
    if drop_determiners is True:
        ents = (
            ent
            if ent[0].pos != DET
            else Span(
                ent.doc, ent.start + 1, ent.end, label=ent.label, vector=ent.vector
            )
            for ent in ents
        )
    if min_freq > 1:
        ents = list(ents)
        freqs = itertoolz.frequencies(ent.lower_ for ent in ents)
        ents = (ent for ent in ents if freqs[ent.lower_] >= min_freq)

    for ent in ents:
        yield ent


def _parse_ent_types(
    ent_types: Optional[Union[str, Set[str]]],
    which: str,
) -> Optional[Union[str, Set[str]]]:
    if not ent_types:
        return None
    elif isinstance(ent_types, str):
        ent_types = ent_types.upper()
        # replace the shorthand numeric case by its corresponding constant
        if ent_types == "NUMERIC":
            return constants.NUMERIC_ENT_TYPES
        else:
            return ent_types
    elif isinstance(ent_types, (set, frozenset, list, tuple)):
        ent_types = {ent_type.upper() for ent_type in ent_types}
        # again, replace the shorthand numeric case by its corresponding constant
        # and include it in the set in case other types are specified
        if any(ent_type == "NUMERIC" for ent_type in ent_types):
            return ent_types.union(constants.NUMERIC_ENT_TYPES)
        else:
            return ent_types
    else:
        allowed_types = (None, str, set, frozenset, list, tuple)
        raise TypeError(
            "{}_types = {} is {}, which is not one of the allowed types: {}".format(
                which, ent_types, type(ent_types), allowed_types
            )
        )


def noun_chunks(
    doc: Doc,
    *,
    drop_determiners: bool = True,
    min_freq: int = 1,
) -> Iterable[Span]:
    """
    Extract an ordered sequence of noun chunks from a spacy-parsed doc, optionally
    filtering by frequency and dropping leading determiners.

    Args:
        doc
        drop_determiners: Remove leading determiners (e.g. "the")
            from phrases (e.g. "the quick brown fox" => "quick brown fox")
        min_freq: Remove chunks that occur in ``doc`` fewer than ``min_freq`` times

    Yields:
        Next noun chunk from ``doc`` in order of appearance in the document
    """
    ncs = doc.noun_chunks
    if drop_determiners is True:
        ncs = (nc if nc[0].pos != DET else nc[1:] for nc in ncs)
    if min_freq > 1:
        ncs = list(ncs)
        freqs = itertoolz.frequencies(nc.lower_ for nc in ncs)
        ncs = (nc for nc in ncs if freqs[nc.lower_] >= min_freq)

    for nc in ncs:
        yield nc


def pos_regex_matches(doc: Union[Doc, Span], pattern: str) -> Iterable[Span]:
    """
    Extract sequences of consecutive tokens from a spacy-parsed doc whose
    part-of-speech tags match the specified regex pattern.

    Args:
        doc
        pattern: Pattern of consecutive POS tags whose corresponding words
            are to be extracted, inspired by the regex patterns used in NLTK's
            `nltk.chunk.regexp`. Tags are uppercase, from the universal tag set;
            delimited by < and >, which are basically converted to parentheses
            with spaces as needed to correctly extract matching word sequences;
            white space in the input doesn't matter.

            Examples (see ``constants.POS_REGEX_PATTERNS``):

            * noun phrase: r'<DET>? (<NOUN>+ <ADP|CONJ>)* <NOUN>+'
            * compound nouns: r'<NOUN>+'
            * verb phrase: r'<VERB>?<ADV>*<VERB>+'
            * prepositional phrase: r'<PREP> <DET>? (<NOUN>+<ADP>)* <NOUN>+'

    Yields:
        Next span of consecutive tokens from ``doc`` whose parts-of-speech match ``pattern``,
        in order of appearance

    Warning:
        *DEPRECATED!* For similar but more powerful and performant functionality,
        use :func:`textacy.extract.matches()` instead.
    """
    utils.deprecated(
        "`pos_regex_matches()` has been deprecated! "
        "for similar but more powerful and performant functionality, "
        "use `textacy.extract.matches()` instead.",
        action="once",
    )
    # standardize and transform the regular expression pattern...
    pattern = re.sub(r"\s", "", pattern)
    pattern = re.sub(r"<([A-Z]+)\|([A-Z]+)>", r"( (\1|\2))", pattern)
    pattern = re.sub(r"<([A-Z]+)>", r"( \1)", pattern)

    tags = " " + " ".join(tok.pos_ for tok in doc)

    for m in re.finditer(pattern, tags):
        yield doc[tags[0 : m.start()].count(" ") : tags[0 : m.end()].count(" ")]


def matches(
    doc: Doc,
    patterns: Union[str, List[str], List[dict], List[List[dict]]],
    *,
    on_match: Callable = None,
) -> Iterable[Span]:
    """
    Extract ``Span`` s from a ``Doc`` matching one or more patterns
    of per-token attr:value pairs, with optional quantity qualifiers.

    Args:
        doc
        patterns:
            One or multiple patterns to match against ``doc``
            using a :class:`spacy.matcher.Matcher`.

            If List[dict] or List[List[dict]], each pattern is specified
            as attr: value pairs per token, with optional quantity qualifiers:

            * ``[{"POS": "NOUN"}]`` matches singular or plural nouns,
              like "friend" or "enemies"
            * ``[{"POS": "PREP"}, {"POS": "DET", "OP": "?"}, {"POS": "ADJ", "OP": "?"}, {"POS": "NOUN", "OP": "+"}]``
              matches prepositional phrases, like "in the future" or "from the distant past"
            * ``[{"IS_DIGIT": True}, {"TAG": "NNS"}]`` matches numbered plural nouns,
              like "60 seconds" or "2 beers"
            * ``[{"POS": "PROPN", "OP": "+"}, {}]`` matches proper nouns and
              whatever word follows them, like "Burton DeWilde yaaasss"

            If str or List[str], each pattern is specified as one or more
            per-token patterns separated by whitespace where attribute, value,
            and optional quantity qualifiers are delimited by colons. Note that
            boolean and integer values have special syntax --- "bool(val)" and
            "int(val)", respectively --- and that wildcard tokens still need
            a colon between the (empty) attribute and value strings.

            * ``"POS:NOUN"`` matches singular or plural nouns
            * ``"POS:PREP POS:DET:? POS:ADJ:? POS:NOUN:+"`` matches prepositional phrases
            * ``"IS_DIGIT:bool(True) TAG:NNS"`` matches numbered plural nouns
            * ``"POS:PROPN:+ :"`` matches proper nouns and whatever word follows them

            Also note that these pattern strings don't support spaCy v2.1's
            "extended" pattern syntax; if you need such complex patterns, it's
            probably better to use a List[dict] or List[List[dict]], anyway.

        on_match: Callback function to act on matches.
            Takes the arguments ``matcher``, ``doc``, ``i`` and ``matches``.

    Yields:
        Next matching ``Span`` in ``doc``, in order of appearance

    Raises:
        TypeError
        ValueError

    See Also:
        * https://spacy.io/usage/rule-based-matching
        * https://spacy.io/api/matcher
    """
    if isinstance(patterns, str):
        patterns = [_make_pattern_from_string(patterns)]
    elif isinstance(patterns, (list, tuple)):
        if all(isinstance(item, str) for item in patterns):
            patterns = [_make_pattern_from_string(pattern) for pattern in patterns]
        elif all(isinstance(item, dict) for item in patterns):
            patterns = [patterns]
        elif all(isinstance(item, (list, tuple)) for item in patterns):
            pass  # already in the right format!
        else:
            raise TypeError(
                "patterns={} is invalid; values must be one of {}".format(
                    patterns, {"str", "List[str]", "List[dict]", "List[list[dict]]"})
            )
    else:
        raise TypeError(
            "patterns={} is invalid; values must be one of {}".format(
                patterns, {"str", "List[str]", "List[dict]", "List[list[dict]]"})
        )
    matcher = Matcher(doc.vocab)
    matcher.add("match", on_match, *patterns)
    for _, start, end in matcher(doc):
        yield doc[start:end]


def _make_pattern_from_string(patstr: str) -> List[dict]:
    """
    Args:
        patstr (str)

    Returns:
        List[dict]
    """
    pattern = []
    for tokpatstr in constants.RE_MATCHER_TOKPAT_DELIM.split(patstr):
        parts = tokpatstr.split(":")
        if 2 <= len(parts) <= 3:
            attr = parts[0]
            attr_val = parts[1]
            if attr and attr_val:
                # handle special bool and int attribute values
                special_val = constants.RE_MATCHER_SPECIAL_VAL.match(attr_val)
                if special_val:
                    attr_val = eval(special_val.group(0))
                tokpat = {attr: attr_val}
            # handle wildcard tokens
            else:
                tokpat = {}
            # handle quantifier ops
            try:
                op_val = parts[2]
                if op_val in constants.MATCHER_VALID_OPS:
                    tokpat["OP"] = op_val
                else:
                    raise ValueError(
                        "op={} invalid; valid choices are {}".format(
                            op_val, constants.MATCHER_VALID_OPS)
                    )
            except IndexError:
                pass
            pattern.append(tokpat)
        else:
            raise ValueError(
                "pattern string '{}' is invalid; "
                "each element in a pattern string must contain an attribute, "
                "a corresponding value, and an optional quantity qualifier, "
                "delimited by colons, like attr:value:op".format(patstr)
            )
    return pattern


def subject_verb_object_triples(doc: Union[Doc, Span]) -> Iterable[Tuple[Span, Span, Span]]:
    """
    Extract an ordered sequence of subject-verb-object (SVO) triples from a
    spacy-parsed doc. Note that this only works for SVO languages.

    Args:
        doc

    Yields:
        Next 3-tuple of spans from ``doc`` representing a (subject, verb, object) triple,
        in order of appearance
    """
    # TODO: What to do about questions, where it may be VSO instead of SVO?
    # TODO: What about non-adjacent verb negations?
    # TODO: What about object (noun) negations?
    if isinstance(doc, Span):
        sents = [doc]
    else:  # spacy.Doc
        sents = doc.sents

    for sent in sents:
        start_i = sent[0].i

        verbs = spacy_utils.get_main_verbs_of_sent(sent)
        for verb in verbs:
            subjs = spacy_utils.get_subjects_of_verb(verb)
            if not subjs:
                continue
            objs = spacy_utils.get_objects_of_verb(verb)
            if not objs:
                continue

            # add adjacent auxiliaries to verbs, for context
            # and add compounds to compound nouns
            verb_span = spacy_utils.get_span_for_verb_auxiliaries(verb)
            verb = sent[verb_span[0] - start_i : verb_span[1] - start_i + 1]
            for subj in subjs:
                subj = sent[
                    spacy_utils.get_span_for_compound_noun(subj)[0]
                    - start_i : subj.i
                    - start_i
                    + 1
                ]
                for obj in objs:
                    if obj.pos == NOUN:
                        span = spacy_utils.get_span_for_compound_noun(obj)
                    elif obj.pos == VERB:
                        span = spacy_utils.get_span_for_verb_auxiliaries(obj)
                    else:
                        span = (obj.i, obj.i)
                    obj = sent[span[0] - start_i : span[1] - start_i + 1]

                    yield (subj, verb, obj)


def acronyms_and_definitions(
    doc: Union[Doc, Span],
    known_acro_defs: Optional[Dict[str, str]] = None,
) -> Dict[str, List[str]]:
    """
    Extract a collection of acronyms and their most likely definitions, if available,
    from a spacy-parsed doc. If multiple definitions are found for a given acronym,
    only the most frequently occurring definition is returned.

    Args:
        doc
        known_acro_defs: If certain acronym/definition pairs
            are known, pass them in as {acronym (str): definition (str)};
            algorithm will not attempt to find new definitions

    Returns:
        Unique acronyms (keys) with matched definitions (values)

    References:
        Taghva, Kazem, and Jeff Gilbreth. "Recognizing acronyms and their definitions."
        International Journal on Document Analysis and Recognition 1.4 (1999): 191-198.
    """
    # process function arguments
    acro_defs: Dict[str, List[Tuple[str, float]]] = collections.defaultdict(list)
    if not known_acro_defs:
        known_acronyms = set()
    else:
        for acro, def_ in known_acro_defs.items():
            acro_defs[acro] = [(def_, 1.0)]
        known_acronyms = set(acro_defs.keys())

    if isinstance(doc, Span):
        sents = [doc]
    else:  # spacy.Doc
        sents = doc.sents

    # iterate over sentences and their tokens
    for sent in sents:
        max_ind = len(sent) - 1

        for i, token in enumerate(sent):

            token_ = token.text
            if token_ in known_acronyms or text_utils.is_acronym(token_) is False:
                continue

            # define definition search window(s)
            window_size = min(2 * len(token_), len(token_) + 5)
            windows = [
                sent[max(i - window_size, 0) : i],
                sent[min(i + 1, max_ind) : min(i + window_size + 1, max_ind)],
            ]
            # if candidate inside (X) or -X-, only look in pre-window
            if 0 < i < max_ind:
                adjacent_tokens = sent[i - 1].text + sent[i + 1].text
                if adjacent_tokens in {"()", "--", "––"}:
                    windows.pop()

            # iterate over possible windows
            # filtering for valid definition strings
            for window in windows:
                window_ = window.text
                # window text can't be all uppercase
                if window_.isupper():
                    continue
                # window can't contain separating punctuation
                if "!" in window_ or "?" in window_ or ":" in window_ or ";" in window_:
                    continue
                # acronym definition can't contain itself: no ouroboros!
                if token_ in window_:
                    continue
                # window must contain at least one character used in acronym
                if not any(char in window_ for char in token_):
                    continue
                definition, confidence = _get_acronym_definition(
                    token_, window, threshold=0.8
                )
                if definition:
                    acro_defs[token_].append((definition, confidence))

            if not acro_defs.get(token_):
                acro_defs[token_].append(("", 0.0))

    # vote by confidence score in the case of multiple definitions
    for acro, defs in acro_defs.items():
        if len(defs) == 1:
            acro_defs[acro] = defs[0][0]
        else:
            acro_defs[acro] = sorted(defs, key=operator.itemgetter(1), reverse=True)[0][
                0
            ]

    return dict(acro_defs)


def _get_acronym_definition(
    acronym: str,
    window: Span,
    threshold: float = 0.8,
) -> Tuple[str, float]:
    """
    Identify most likely definition for an acronym given a list of tokens.

    Args:
        acronym: acronym for which definition is sought
        window: a span of tokens from which definition extraction will be attempted
        threshold: minimum "confidence" in definition required for acceptance;
            valid values in [0.0, 1.0]; higher value => stricter threshold

    Returns:
        Most likely definition for given acronym ('' if none found),
        along with the confidence assigned to it

    References:
        Taghva, Kazem, and Jeff Gilbreth. "Recognizing acronyms and their definitions."
        International Journal on Document Analysis and Recognition 1.4 (1999): 191-198.
    """

    def build_lcs_matrix(X, Y):
        m = len(X)
        n = len(Y)
        b = np.zeros((m, n), dtype=int)
        c = np.zeros((m, n), dtype=int)
        for i in range(0, m):
            for j in range(0, n):
                if X[i] == Y[j]:
                    c[i, j] = c[i - 1, j - 1] + 1
                    b[i, j] = 1
                elif c[i - 1, j] >= c[i, j - 1]:
                    c[i, j] = c[i - 1, j]
                else:
                    c[i, j] = c[i, j - 1]
        return c, b

    def parse_lcs_matrix(b, start_i, start_j, lcs_length, stack, vectors):
        m = b.shape[0]
        n = b.shape[1]
        for i in range(start_i, m):
            for j in range(start_j, n):
                if b[i, j] == 1:
                    s = (i, j)
                    stack.append(s)
                    if lcs_length == 1:
                        vec = [np.NaN] * n
                        for k, l in stack:
                            vec[l] = k
                        vectors.append(vec)
                    else:
                        parse_lcs_matrix(
                            b, i + 1, j + 1, lcs_length - 1, stack, vectors
                        )
                    stack = []
        return vectors

    def vector_values(v, types):
        vv = {}
        first = v.index(int(np.nanmin(v)))
        last = v.index(int(np.nanmax(v)))
        vv["size"] = (last - first) + 1
        vv["distance"] = len(v) - last
        vv["stop_count"] = 0
        vv["misses"] = 0
        for i in range(first, last + 1):
            if v[i] >= 0 and types[i] == "s":
                vv["stop_count"] += 1
            elif v[i] is None and types[i] not in ["s", "h"]:
                vv["misses"] += 1
        return vv

    def compare_vectors(A, B, types):
        vv_A = vector_values(A, types)
        vv_B = vector_values(B, types)
        # no one-letter matches, sorryboutit
        if vv_A["size"] == 1:
            return B
        elif vv_B["size"] == 1:
            return A
        if vv_A["misses"] > vv_B["misses"]:
            return B
        elif vv_A["misses"] < vv_B["misses"]:
            return A
        if vv_A["stop_count"] > vv_B["stop_count"]:
            return B
        if vv_A["stop_count"] < vv_B["stop_count"]:
            return A
        if vv_A["distance"] > vv_B["distance"]:
            return B
        elif vv_A["distance"] < vv_B["distance"]:
            return A
        if vv_A["size"] > vv_B["size"]:
            return B
        elif vv_A["size"] < vv_B["size"]:
            return A
        return A

    # get definition window's leading characters and word types
    def_leads = []
    def_types = []
    for tok in window:
        tok_text = tok.text
        if tok.is_stop:
            def_leads.append(tok_text[0])
            def_types.append("s")
        elif text_utils.is_acronym(tok_text):
            def_leads.append(tok_text[0])
            def_types.append("a")
        elif "-" in tok_text and not tok_text.startswith("-"):
            tok_split = [t[0] for t in tok_text.split("-") if t]
            def_leads.extend(tok_split)
            def_types.extend(
                "H" if i == 0 else "h" for i in range(len(tok_split))
            )
        else:
            def_leads.append(tok_text[0])
            def_types.append("w")
    def_leads = "".join(def_leads).lower()
    def_types = "".join(def_types)

    # extract alphanumeric characters from acronym
    acr_leads = "".join(c for c in acronym if c.isalnum())
    # handle special cases of '&' and trailing 's'
    acr_leads = acr_leads.replace("&", "a")
    if acr_leads.endswith("s"):
        # bail out if it's only a 2-letter acronym to start with, e.g. 'Is'
        if len(acr_leads) == 2:
            return ("", 0)
        acr_leads = acr_leads[:-1]
    acr_leads = acr_leads.lower()

    c, b = build_lcs_matrix(acr_leads, def_leads)

    # 4.4.1
    lcs_length = c[c.shape[0] - 1, c.shape[1] - 1]
    confidence = lcs_length / len(acronym)
    if confidence < threshold:
        return ("", confidence)

    vecs = parse_lcs_matrix(b, 0, 0, lcs_length, [], [])
    # first letter of acronym must be present
    vecs = [vec for vec in vecs if 0 in vec]
    if not vecs:
        return ("", confidence)

    best_vec = vecs[0]
    for vec in vecs[1:]:
        best_vec = compare_vectors(best_vec, vec, def_types)

    first = best_vec.index(int(np.nanmin(best_vec)))
    last = best_vec.index(int(np.nanmax(best_vec)))

    definition = window[first : last + 1].text
    if len(definition.split()) == 1:
        return ("", confidence)

    return (definition, confidence)


def semistructured_statements(
    doc: Doc,
    entity: str,
    *,
    cue: str = "be",
    ignore_entity_case: bool = True,
    min_n_words: int = 1,
    max_n_words: int = 20,
):
    """
    Extract "semi-structured statements" from a spacy-parsed doc, each as a
    (entity, cue, fragment) triple. This is similar to subject-verb-object triples.

    Args:
        doc
        entity: a noun or noun phrase of some sort (e.g. "President Obama",
            "global warming", "Python")
        cue: verb lemma with which ``entity`` is associated
            (e.g. "talk about", "have", "write")
        ignore_entity_case: If True, entity matching is case-independent
        min_n_words: Min number of tokens allowed in a matching fragment
        max_n_words: Max number of tokens allowed in a matching fragment

    Yields:
        (:class:`spacy.tokens.Span` or :class:`spacy.tokens.Token`, :class:`spacy.tokens.Span` or :class:`spacy.tokens.Token`, :class:`spacy.tokens.Span`):
        where each element is a matching (entity, cue, fragment) triple

    Notes:
        Inspired by N. Diakopoulos, A. Zhang, A. Salway. Visual Analytics of
        Media Frames in Online News and Blogs. IEEE InfoVis Workshop on Text
        Visualization. October, 2013.

        Which itself was inspired by by Salway, A.; Kelly, L.; Skadiņa, I.; and
        Jones, G. 2010. Portable Extraction of Partially Structured Facts from
        the Web. In Proc. ICETAL 2010, LNAI 6233, 345-356. Heidelberg, Springer.
    """
    if ignore_entity_case is True:
        entity_toks = entity.lower().split(" ")
        get_tok_text = lambda x: x.lower_
    else:
        entity_toks = entity.split(" ")
        get_tok_text = lambda x: x.text
    first_entity_tok = entity_toks[0]
    n_entity_toks = len(entity_toks)
    cue = cue.lower()
    cue_toks = cue.split(" ")
    n_cue_toks = len(cue_toks)

    def is_good_last_tok(tok):
        if tok.is_punct:
            return False
        if tok.pos in {CONJ, DET}:
            return False
        return True

    for sent in doc.sents:
        for tok in sent:

            # filter by entity
            if get_tok_text(tok) != first_entity_tok:
                continue
            if n_entity_toks == 1:
                the_entity = tok
                the_entity_root = the_entity
            if tok.i + n_cue_toks >= len(doc):
                continue
            elif all(
                get_tok_text(tok.nbor(i=i + 1)) == et
                for i, et in enumerate(entity_toks[1:])
            ):
                the_entity = doc[tok.i : tok.i + n_entity_toks]
                the_entity_root = the_entity.root
            else:
                continue

            # filter by cue
            terh = the_entity_root.head
            if terh.lemma_ != cue_toks[0]:
                continue
            if n_cue_toks == 1:
                min_cue_i = terh.i
                max_cue_i = terh.i + n_cue_toks
                the_cue = terh
            elif all(
                terh.nbor(i=i + 1).lemma_ == ct for i, ct in enumerate(cue_toks[1:])
            ):
                min_cue_i = terh.i
                max_cue_i = terh.i + n_cue_toks
                the_cue = doc[terh.i : max_cue_i]
            else:
                continue
            if the_entity_root in the_cue.rights:
                continue

            # now add adjacent auxiliary and negating tokens to the cue, for context
            try:
                min_cue_i = min(
                    left.i
                    for left in itertools.takewhile(
                        lambda x: x.dep_ in {"aux", "neg"},
                        reversed(list(the_cue.lefts)),
                    )
                )
            except ValueError:
                pass
            try:
                max_cue_i = max(
                    right.i
                    for right in itertools.takewhile(
                        lambda x: x.dep_ in {"aux", "neg"}, the_cue.rights
                    )
                )
            except ValueError:
                pass
            if max_cue_i - min_cue_i > 1:
                the_cue = doc[min_cue_i:max_cue_i]
            else:
                the_cue = doc[min_cue_i]

            # filter by fragment
            try:
                min_frag_i = min(right.left_edge.i for right in the_cue.rights)
                max_frag_i = max(right.right_edge.i for right in the_cue.rights)
            except ValueError:
                continue
            while is_good_last_tok(doc[max_frag_i]) is False:
                max_frag_i -= 1
            n_fragment_toks = max_frag_i - min_frag_i
            if (
                n_fragment_toks <= 0
                or n_fragment_toks < min_n_words
                or n_fragment_toks > max_n_words
            ):
                continue
            # HACK...
            if min_frag_i == max_cue_i - 1:
                min_frag_i += 1
            the_fragment = doc[min_frag_i : max_frag_i + 1]

            yield (the_entity, the_cue, the_fragment)


def direct_quotations(doc: Doc) -> Iterable[Tuple[Span, Token, Span]]:
    """
    Baseline, not-great attempt at direction quotation extraction (no indirect
    or mixed quotations) using rules and patterns. English only.

    Args:
        doc (:class:`spacy.tokens.Doc`)

    Yields:
        Next quotation in ``doc`` represented as a (speaker, reporting verb, quotation) 3-tuple

    Notes:
        Loosely inspired by Krestel, Bergler, Witte. "Minding the Source: Automatic
        Tagging of Reported Speech in Newspaper Articles".

    TODO: Better approach would use ML, but needs a training dataset.
    """
    doc_lang = doc.vocab.lang
    if doc_lang != "en":
        raise NotImplementedError("sorry, English-language texts only :(")
    quote_end_punct = {",", ".", "?", "!"}
    quote_indexes = set(
        itertoolz.concat(
            (m.start(), m.end() - 1)
            for m in re.finditer(r"(\".*?\")|(''.*?'')|(``.*?'')", doc.text)
        )
    )
    quote_positions = list(
        itertoolz.partition(2, sorted(tok.i for tok in doc if tok.idx in quote_indexes))
    )
    sents = list(doc.sents)
    sent_positions = [(sent.start, sent.end) for sent in sents]

    for q0, q1 in quote_positions:
        quote = doc[q0 : q1 + 1]

        # we're only looking for direct quotes, not indirect or mixed
        if not any(char in quote_end_punct for char in quote.text[-4:]):
            continue

        # get adjacent sentences
        candidate_sent_indexes = []
        for i, (s0, s1) in enumerate(sent_positions):

            if s0 <= q1 + 1 and s1 > q1:
                candidate_sent_indexes.append(i)
            elif s0 < q0 and s1 >= q0 - 1:
                candidate_sent_indexes.append(i)

        for si in candidate_sent_indexes:
            sent = sents[si]

            # get any reporting verbs
            rvs = [
                tok
                for tok in sent
                if spacy_utils.preserve_case(tok) is False
                and tok.lemma_ in constants.REPORTING_VERBS
                and tok.pos_ == "VERB"
                and not any(oq0 <= tok.i <= oq1 for oq0, oq1 in quote_positions)
            ]

            # get target offset against which to measure distances of NEs
            if rvs:
                if len(rvs) == 1:
                    rv = rvs[0]
                else:
                    min_rv_dist = 1000
                    for rv_candidate in rvs:
                        rv_dist = min(abs(rv_candidate.i - qp) for qp in (q0, q1))
                        if rv_dist < min_rv_dist:
                            rv = rv_candidate
                            min_rv_dist = rv_dist
                        else:
                            break
            else:
                # TODO: do we have no other recourse?!
                continue

            try:
                # rv_subj = _find_subjects(rv)[0]
                rv_subj = spacy_utils.get_subjects_of_verb(rv)[0]
            except IndexError:
                continue
            #         if rv_subj.text in {'he', 'she'}:
            #             for ne in entities(doc, include_types={'PERSON'}):
            #                 if ne.start < rv_subj.i:
            #                     speaker = ne
            #                 else:
            #                     break
            #         else:
            span = spacy_utils.get_span_for_compound_noun(rv_subj)
            speaker = doc[span[0] : span[1] + 1]

            yield (speaker, rv, quote)
            break



import re
import sys
import unicodedata
from typing import Dict, Pattern

# compile regexes, so we don't do this on the fly and rely on caching

RE_LINEBREAK: Pattern = re.compile(r"(\r\n|[\n\v])+")
RE_NONBREAKING_SPACE: Pattern = re.compile(r"[^\S\n\v]+", flags=re.UNICODE)
RE_ZWSP: Pattern = re.compile(r"[\u200B\u2060\uFEFF]+")

# source: https://gist.github.com/dperini/729294
RE_URL: Pattern = re.compile(
    r"(?:^|(?<![\w/.]))"
    # protocol identifier
    # r"(?:(?:https?|ftp)://)"  <-- alt?
    r"(?:(?:https?://|ftp://|www\d{0,3}\.))"
    # user:pass authentication
    r"(?:\S+(?::\S*)?@)?"
    r"(?:"
    # IP address exclusion
    # private & local networks
    r"(?!(?:10|127)(?:\.\d{1,3}){3})"
    r"(?!(?:169\.254|192\.168)(?:\.\d{1,3}){2})"
    r"(?!172\.(?:1[6-9]|2\d|3[0-1])(?:\.\d{1,3}){2})"
    # IP address dotted notation octets
    # excludes loopback network 0.0.0.0
    # excludes reserved space >= 224.0.0.0
    # excludes network & broadcast addresses
    # (first & last IP address of each class)
    r"(?:[1-9]\d?|1\d\d|2[01]\d|22[0-3])"
    r"(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5])){2}"
    r"(?:\.(?:[1-9]\d?|1\d\d|2[0-4]\d|25[0-4]))"
    r"|"
    # host name
    r"(?:(?:[a-z\u00a1-\uffff0-9]-?)*[a-z\u00a1-\uffff0-9]+)"
    # domain name
    r"(?:\.(?:[a-z\u00a1-\uffff0-9]-?)*[a-z\u00a1-\uffff0-9]+)*"
    # TLD identifier
    r"(?:\.(?:[a-z\u00a1-\uffff]{2,}))"
    r")"
    # port number
    r"(?::\d{2,5})?"
    # resource path
    r"(?:/\S*)?"
    r"(?:$|(?![\w?!+&/]))",
    flags=re.UNICODE | re.IGNORECASE)

RE_SHORT_URL: Pattern = re.compile(
    r"(?:^|(?<![\w/.]))"
    # optional scheme
    r"(?:(?:https?://)?)"
    # domain
    r"(?:\w-?)*?\w+(?:\.[a-z]{2,12}){1,3}"
    r"/"
    # hash
    r"[^\s.,?!'\"|+]{2,12}"
    r"(?:$|(?![\w?!+&/]))",
    flags=re.UNICODE | re.IGNORECASE)

RE_EMAIL: Pattern = re.compile(
    r"(?:mailto:)?"
    r"(?:^|(?<=[^\w@.)]))([\w+-](\.(?!\.))?)*?[\w+-]@(?:\w-?)*?\w+(\.([a-z]{2,})){1,3}"
    r"(?:$|(?=\b))",
    flags=re.UNICODE | re.IGNORECASE)

RE_USER_HANDLE: Pattern = re.compile(
    r"(?:^|(?<![\w@.]))@\w+",
    flags=re.UNICODE | re.IGNORECASE)

RE_HASHTAG: Pattern = re.compile(
    r"(?:^|(?<![\w#＃.]))(#|＃)(?!\d)\w+",
    flags=re.UNICODE | re.IGNORECASE)

RE_PHONE_NUMBER: Pattern = re.compile(
    # core components of a phone number
    r"(?:^|(?<=[^\w)]))(\+?1[ .-]?)?(\(?\d{3}\)?[ .-]?)?(\d{3}[ .-]?\d{4})"
    # extensions, etc.
    r"(\s?(?:ext\.?|[#x-])\s?\d{2,6})?(?:$|(?=\W))",
    flags=re.UNICODE | re.IGNORECASE)

RE_NUMBER: Pattern = re.compile(
    r"(?:^|(?<=[^\w,.]))[+–-]?"
    r"(([1-9]\d{0,2}(,\d{3})+(\.\d*)?)|([1-9]\d{0,2}([ .]\d{3})+(,\d*)?)|(\d*?[.,]\d+)|\d+)"
    r"(?:$|(?=\b))")

RE_CURRENCY_SYMBOL: Pattern = re.compile(
    r"[$¢£¤¥ƒ֏؋৲৳૱௹฿៛ℳ元円圆圓﷼\u20A0-\u20C0]",
    flags=re.UNICODE)

RE_EMOJI: Pattern
if sys.maxunicode < 0x10ffff:
    RE_EMOJI = re.compile(
        r"[\u2600-\u26FF\u2700-\u27BF]",
        flags=re.UNICODE | re.IGNORECASE)
else:
    RE_EMOJI = re.compile(
        r"[\u2600-\u26FF\u2700-\u27BF\U0001F300-\U0001F5FF\U0001F600-\U0001F64F\U0001F680-\U0001F6FF\U0001F900-\U0001F9FF\U0001FA70-\U0001FAFF]",
        flags=re.UNICODE | re.IGNORECASE)

RE_HYPHENATED_WORD: Pattern = re.compile(
    r"(\w{2,}(?<!\d))\-\s+((?!\d)\w{2,})",
    flags=re.UNICODE | re.IGNORECASE)


# build mapping of unicode punctuation symbol ordinals to their replacements
# and lazy-load the big one, since it's relatively expensive to compute

PUNCT_TRANSLATION_TABLE = None


def _get_punct_translation_table():
    global PUNCT_TRANSLATION_TABLE
    if PUNCT_TRANSLATION_TABLE is None:
        PUNCT_TRANSLATION_TABLE = dict.fromkeys(
            (i for i in range(sys.maxunicode)
             if unicodedata.category(chr(i)).startswith("P")),
            " "
        )
    return PUNCT_TRANSLATION_TABLE


QUOTE_TRANSLATION_TABLE: Dict[int, int] = {
    ord(x): ord(y)
    for x, y in zip("ʼ‘’´`“”", "'''''\"\"")
}



import numpy as np

try:
    import matplotlib.pyplot as plt
except ImportError:
    pass


RC_PARAMS = {
    "axes.axisbelow": True,
    "axes.edgecolor": ".8",
    "axes.facecolor": "white",
    "axes.grid": True,
    "axes.labelcolor": ".15",
    "axes.linewidth": 1.0,
    "figure.facecolor": "white",
    "font.family": ["sans-serif"],
    "font.sans-serif": ["Arial", "Liberation Sans", "sans-serif"],
    "grid.color": ".8",
    "grid.linestyle": "-",
    "image.cmap": "Greys",
    "legend.frameon": False,
    "legend.numpoints": 1,
    "legend.scatterpoints": 1,
    "lines.solid_capstyle": "round",
    "text.color": ".15",
    "xtick.color": ".15",
    "xtick.direction": "out",
    "xtick.major.size": 0.0,
    "xtick.minor.size": 0.0,
    "ytick.color": ".15",
    "ytick.direction": "out",
    "ytick.major.size": 0.0,
    "ytick.minor.size": 0.0,
}

COLOR_PAIRS = (
    (
        (0.65098041296005249, 0.80784314870834351, 0.89019608497619629),
        (0.12572087695201239, 0.47323337360924367, 0.707327968232772),
    ),
    (
        (0.68899655751153521, 0.8681737867056154, 0.54376011946622071),
        (0.21171857311445125, 0.63326415104024547, 0.1812226118410335),
    ),
    (
        (0.98320646005518297, 0.5980161709820524, 0.59423301088459368),
        (0.89059593116535862, 0.10449827132271793, 0.11108035462744099),
    ),
    (
        (0.99175701702342312, 0.74648213716698619, 0.43401768935077328),
        (0.99990772780250103, 0.50099192647372981, 0.0051211073118098693),
    ),
    (
        (0.78329874347238004, 0.68724338552531095, 0.8336793640080622),
        (0.42485198495434734, 0.2511495584950722, 0.60386007743723258),
    ),
    (
        (0.99760092286502611, 0.99489427150464516, 0.5965244373854468),
        (0.69411766529083252, 0.3490196168422699, 0.15686275064945221),
    ),
)


def draw_termite_plot(
    values_mat,
    col_labels,
    row_labels,
    *,
    highlight_cols=None,
    highlight_colors=None,
    save=False,
    rc_params=None,
):
    """
    Make a "termite" plot, typically used for assessing topic models with a tabular
    layout that promotes comparison of terms both within and across topics.

    Args:
        values_mat (:class:`np.ndarray` or matrix): matrix of values with shape
            (# row labels, # col labels) used to size the dots on the grid
        col_labels (seq[str]): labels used to identify x-axis ticks on the grid
        row_labels(seq[str]): labels used to identify y-axis ticks on the grid
        highlight_cols (int or seq[int], optional): indices for columns
            to visually highlight in the plot with contrasting colors
        highlight_colors (tuple of 2-tuples): each 2-tuple corresponds to a pair
            of (light/dark) matplotlib-friendly colors used to highlight a single
            column; if not specified (default), a good set of 6 pairs are used
        save (str, optional): give the full /path/to/fname on disk to save figure
        rc_params (dict, optional): allow passing parameters to rc_context in matplotlib.plyplot,
            details in https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.pyplot.rc_context.html

    Returns:
        :obj:`matplotlib.axes.Axes.axis`: Axis on which termite plot is plotted.

    Raises:
        ValueError: if more columns are selected for highlighting than colors
            or if any of the inputs' dimensions don't match

    References:
        Chuang, Jason, Christopher D. Manning, and Jeffrey Heer. "Termite:
        Visualization techniques for assessing textual topic models."
        Proceedings of the International Working Conference on Advanced
        Visual Interfaces. ACM, 2012.

    See Also:
        :meth:`TopicModel.termite_plot() <textacy.tm.topic_model.TopicModel.termite_plot>`
    """
    try:
        plt
    except NameError:
        raise ImportError(
            "`matplotlib` is not installed, so `textacy.viz` won't work; "
            "install it individually via `$ pip install matplotlib`, or "
            "along with textacy via `pip install textacy[viz]`."
        )
    n_rows, n_cols = values_mat.shape
    max_val = np.max(values_mat)

    if n_rows != len(row_labels):
        msg = "values_mat and row_labels dimensions don't match: {} vs. {}".format(
            n_rows, len(row_labels)
        )
        raise ValueError(msg)
    if n_cols != len(col_labels):
        msg = "values_mat and col_labels dimensions don't match: {} vs. {}".format(
            n_cols, len(col_labels)
        )
        raise ValueError(msg)

    if highlight_colors is None:
        highlight_colors = COLOR_PAIRS
    if highlight_cols is not None:
        if isinstance(highlight_cols, int):
            highlight_cols = (highlight_cols,)
        elif len(highlight_cols) > len(highlight_colors):
            msg = "no more than {} columns may be highlighted at once".format(
                len(highlight_colors)
            )
            raise ValueError(msg)
        highlight_colors = {hc: COLOR_PAIRS[i] for i, hc in enumerate(highlight_cols)}

    _rc_params = RC_PARAMS.copy()
    if rc_params:
        _rc_params.update(rc_params)

    with plt.rc_context(RC_PARAMS):
        fig, ax = plt.subplots(figsize=(pow(n_cols, 0.8), pow(n_rows, 0.66)))

        _ = ax.set_yticks(range(n_rows))
        yticklabels = ax.set_yticklabels(row_labels, fontsize=14, color="gray")
        if highlight_cols is not None:
            for i, ticklabel in enumerate(yticklabels):
                max_tick_val = max(values_mat[i, hc] for hc in highlight_cols)
                for hc in highlight_cols:
                    if max_tick_val > 0 and values_mat[i, hc] == max_tick_val:
                        ticklabel.set_color(highlight_colors[hc][1])

        ax.get_xaxis().set_ticks_position("top")
        _ = ax.set_xticks(range(n_cols))
        xticklabels = ax.set_xticklabels(
            col_labels, fontsize=14, color="gray", rotation=30, ha="left"
        )
        if highlight_cols is not None:
            gridlines = ax.get_xgridlines()
            for i, ticklabel in enumerate(xticklabels):
                if i in highlight_cols:
                    ticklabel.set_color(highlight_colors[i][1])
                    gridlines[i].set_color(highlight_colors[i][0])
                    gridlines[i].set_alpha(0.5)

        for col_ind in range(n_cols):
            if highlight_cols is not None and col_ind in highlight_cols:
                ax.scatter(
                    [col_ind for _ in range(n_rows)],
                    [i for i in range(n_rows)],
                    s=600 * (values_mat[:, col_ind] / max_val),
                    alpha=0.5,
                    linewidth=1,
                    color=highlight_colors[col_ind][0],
                    edgecolor=highlight_colors[col_ind][1],
                )
            else:
                ax.scatter(
                    [col_ind for _ in range(n_rows)],
                    [i for i in range(n_rows)],
                    s=600 * (values_mat[:, col_ind] / max_val),
                    alpha=0.5,
                    linewidth=1,
                    color="lightgray",
                    edgecolor="gray",
                )

            _ = ax.set_xlim(left=-1, right=n_cols)
            _ = ax.set_ylim(bottom=-1, top=n_rows)

            ax.invert_yaxis()  # otherwise, values/labels go from bottom to top

    if save:
        fig.savefig(save, bbox_inches="tight", dpi=100)

    return ax



import collections
import functools
import itertools
import string
from typing import Iterable, List, NamedTuple, Tuple, Union

from cachetools import cached
from cachetools.keys import hashkey
from spacy.tokens import Doc, Span

from .. import cache, datasets, resources


concept_net = resources.ConceptNet()
udhr = datasets.UDHR()

AugTok = collections.namedtuple("AugTok", ["text", "ws", "pos", "is_word", "syns"])
"""tuple: Minimal token data required for data augmentation transforms."""


def to_aug_toks(spacy_obj: Union[Doc, Span]) -> List[AugTok]:
    """
    Transform a spaCy ``Doc`` or ``Span`` into a list of ``AugTok`` objects,
    suitable for use in data augmentation transform functions.
    """
    if not isinstance(spacy_obj, (Doc, Span)):
        raise TypeError(
            "`spacy_obj` must be of type {}, not {}".format((Doc, Span), type(spacy_obj))
        )
    lang = spacy_obj.vocab.lang
    toks_syns: Iterable[List[str]]
    if concept_net.filepath is None or lang not in concept_net.synonyms:
        toks_syns = ([] for _ in spacy_obj)
    else:
        toks_syns = (
            concept_net.get_synonyms(tok.text, lang=lang, sense=tok.pos_)
            if not (tok.is_punct or tok.is_space) else []
            for tok in spacy_obj
        )
    return [
        AugTok(
            text=tok.text,
            ws=tok.whitespace_,
            pos=tok.pos_,
            is_word=(not (tok.is_punct or tok.is_space)),
            syns=syns,
        )
        for tok, syns in zip(spacy_obj, toks_syns)
    ]


@cached(cache.LRU_CACHE, key=functools.partial(hashkey, "char_weights"))
def get_char_weights(lang: str) -> List[Tuple[str, int]]:
    """
    Get lang-specific character weights for use in certain data augmentation transforms,
    based on texts in :class:`textacy.datasets.UDHR`.

    Args:
        lang: Standard two-letter language code.

    Returns:
        Collection of (character, weight) pairs, based on the distribution of characters
        found in the source text.
    """
    try:
        char_weights = list(
            collections.Counter(
                char
                for text in udhr.texts(lang=lang)
                for char in text
                if char.isalnum()
            ).items()
        )
    except ValueError:
        char_weights = list(
            zip(string.ascii_lowercase + string.digits, itertools.repeat(1))
        )
    return char_weights



"""
Remove
------

Remove aspects of raw text that may be unwanted for certain use cases.
"""
import re
import unicodedata
from typing import Optional

from .resources import _get_punct_translation_table


def remove_accents(text: str, *, fast: bool = False) -> str:
    """
    Remove accents from any accented unicode characters in ``text``, either by
    replacing them with ASCII equivalents or removing them entirely.

    Args:
        text
        fast: If False, accents are removed from any unicode symbol
            with a direct ASCII equivalent; if True, accented chars
            for all unicode symbols are removed, regardless.

            .. note:: ``fast=True`` can be significantly faster than ``fast=False``,
               but its transformation of ``text`` is less "safe" and more likely
               to result in changes of meaning, spelling errors, etc.

    Returns:
        str

    Raises:
        ValueError: If ``method`` is not in {"unicode", "ascii"}.

    See Also:
        For a more powerful (but slower) alternative, check out ``unidecode``:
        https://github.com/avian2/unidecode
    """
    if fast is False:
        return "".join(
            char
            for char in unicodedata.normalize("NFKD", text)
            if not unicodedata.combining(char)
        )
    else:
        return (
            unicodedata.normalize("NFKD", text)
            .encode("ascii", errors="ignore")
            .decode("ascii")
        )


def remove_punctuation(text: str, *, marks: Optional[str] = None) -> str:
    """
    Remove punctuation from ``text`` by replacing all instances of ``marks``
    with whitespace.

    Args:
        text
        marks: Remove only those punctuation marks specified here.
            For example, ",;:" removes commas, semi-colons, and colons.
            If None, *all* unicode punctuation marks are removed.

    Returns:
        str

    Note:
        When ``marks=None``, Python's built-in :meth:`str.translate()` is
        used to remove punctuation; otherwise, a regular expression is used.
        The former's performance is about 5-10x faster.
    """
    if marks:
        return re.sub("[{}]+".format(re.escape(marks)), " ", text, flags=re.UNICODE)
    else:
        return text.translate(_get_punct_translation_table())
